from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

from great_expectations.compatibility.pyspark import (
    functions as F,
)
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.column_aggregate_metric_provider import (
    ColumnAggregateMetricProvider,
    column_aggregate_partial,
    column_aggregate_value,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    import pandas as pd

    from great_expectations.compatibility import pyspark, sqlalchemy
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )


class ColumnDistinctValues(ColumnAggregateMetricProvider):
    metric_name = "column.distinct_values"

    @column_aggregate_value(engine=PandasExecutionEngine)  # type: ignore[misc] # untyped-decorator
    def _pandas(cls, column: pd.Series, **kwargs) -> Set[Any]:
        return set(column.unique())

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: Dict[str, str],
        **kwargs,
    ) -> Set[Any]:
        """
        Past implementations of column.distinct_values depended on column.value_counts.
        This was causing performance issues due to the complex query used in column.value_counts and subsequent
        in-memory operations.
        """  # noqa: E501 # FIXME CoP
        selectable: sqlalchemy.Selectable
        accessor_domain_kwargs: Dict[str, str]
        (
            selectable,
            _,
            accessor_domain_kwargs,
        ) = execution_engine.get_compute_domain(metric_domain_kwargs, MetricDomainTypes.COLUMN)
        column_name: str = accessor_domain_kwargs["column"]
        column: sqlalchemy.ColumnClause = sa.column(column_name)

        distinct_values: List[sqlalchemy.Row]
        if hasattr(column, "is_not"):
            distinct_values = execution_engine.execute_query(  # type: ignore[assignment] # FIXME CoP
                sa.select(column).where(column.is_not(None)).distinct().select_from(selectable)  # type: ignore[arg-type] # FIXME CoP
            ).fetchall()
        else:
            distinct_values = execution_engine.execute_query(  # type: ignore[assignment] # FIXME CoP
                sa.select(column).where(column.isnot(None)).distinct().select_from(selectable)  # type: ignore[arg-type] # FIXME CoP
            ).fetchall()
        # Vectorized operation is not faster here due to overhead of converting to and from numpy array  # noqa: E501 # FIXME CoP
        return {row[0] for row in distinct_values}

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs: Dict[str, str],
        **kwargs,
    ) -> Set[Any]:
        """
        Past implementations of column.distinct_values depended on column.value_counts.
        This was causing performance issues due to the complex query used in column.value_counts and subsequent
        in-memory operations.
        """  # noqa: E501 # FIXME CoP
        df: pyspark.DataFrame
        accessor_domain_kwargs: Dict[str, str]
        (
            df,
            _,
            accessor_domain_kwargs,
        ) = execution_engine.get_compute_domain(metric_domain_kwargs, MetricDomainTypes.COLUMN)
        column_name: str = accessor_domain_kwargs["column"]
        distinct_values: List[pyspark.Row] = (
            df.select(F.col(column_name))
            .distinct()
            .where(F.col(column_name).isNotNull())
            .rdd.flatMap(lambda x: x)
            .collect()
        )
        return set(distinct_values)


class ColumnDistinctValuesCount(ColumnAggregateMetricProvider):
    metric_name = "column.distinct_values.count"

    @column_aggregate_value(engine=PandasExecutionEngine)  # type: ignore[misc] # untyped-decorator
    def _pandas(cls, column: pd.Series, **kwargs) -> int:
        return column.nunique()

    @column_aggregate_partial(engine=SqlAlchemyExecutionEngine)  # type: ignore[misc] # untyped-decorator
    def _sqlalchemy(
        cls,
        column: sqlalchemy.ColumnClause,
        **kwargs,
    ) -> sqlalchemy.Selectable:
        """
        Past implementations of column.distinct_values.count depended on column.value_counts and column.distinct_values.
        This was causing performance issues due to the complex query used in column.value_counts and subsequent
        in-memory operations.
        """  # noqa: E501 # FIXME CoP
        return sa.func.count(sa.distinct(column))

    @column_aggregate_partial(engine=SparkDFExecutionEngine)  # type: ignore[misc] # untyped-decorator
    def _spark(
        cls,
        column: pyspark.Column,
        **kwargs,
    ) -> pyspark.Column:
        """
        Past implementations of column.distinct_values.count depended on column.value_counts and column.distinct_values.
        This was causing performance issues due to the complex query used in column.value_counts and subsequent
        in-memory operations.
        """  # noqa: E501 # FIXME CoP
        return F.countDistinct(column)


class ColumnDistinctValuesCountUnderThreshold(ColumnAggregateMetricProvider):
    metric_name = "column.distinct_values.count.under_threshold"
    condition_keys = ("threshold",)

    @column_aggregate_value(engine=PandasExecutionEngine)  # type: ignore[misc] # untyped-decorator
    def _pandas(cls, column: pd.Series, threshold: int, **kwargs) -> bool:
        return column.nunique() < threshold

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        metric_value_kwargs: Dict[str, int],
        metrics: Dict[str, int],
        **kwargs,
    ) -> bool:
        return metrics["column.distinct_values.count"] < metric_value_kwargs["threshold"]

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        metric_value_kwargs: Dict[str, int],
        metrics: Dict[str, int],
        **kwargs,
    ) -> bool:
        return metrics["column.distinct_values.count"] < metric_value_kwargs["threshold"]

    @classmethod
    @override
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[Dict] = None,
    ):
        """Returns a dictionary of given metric names and their corresponding configuration,
        specifying the metric types and their respective domains"""
        dependencies: dict = super()._get_evaluation_dependencies(
            metric=metric,
            configuration=configuration,
            execution_engine=execution_engine,
            runtime_configuration=runtime_configuration,
        )
        if metric.metric_name == "column.distinct_values.count.under_threshold":
            dependencies["column.distinct_values.count"] = MetricConfiguration(
                metric_name="column.distinct_values.count",
                metric_domain_kwargs=metric.metric_domain_kwargs,
                metric_value_kwargs=None,
            )
        return dependencies



import pandas as pd

from tests.integration.conftest import (
    MultiSourceBatch,
    MultiSourceTestConfig,
    multi_source_batch_setup,
)
from tests.integration.test_utils.data_source_config import (
    DatabricksDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

DATA_FRAME = pd.DataFrame({"a": [1, 2, 3]})


ALL_COMPARISON_TO_BASE_SOURCES = [
    MultiSourceTestConfig(
        comparison=PostgreSQLDatasourceTestConfig(), base=PostgreSQLDatasourceTestConfig()
    ),
    MultiSourceTestConfig(
        comparison=PostgreSQLDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=SnowflakeDatasourceTestConfig(), base=SnowflakeDatasourceTestConfig()
    ),
    MultiSourceTestConfig(
        comparison=SnowflakeDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=DatabricksDatasourceTestConfig(),
        base=DatabricksDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=DatabricksDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=RedshiftDatasourceTestConfig(),
        base=RedshiftDatasourceTestConfig(),
    ),
    MultiSourceTestConfig(
        comparison=RedshiftDatasourceTestConfig(),
        base=SqliteDatasourceTestConfig(),
    ),
]


@multi_source_batch_setup(
    multi_source_test_configs=ALL_COMPARISON_TO_BASE_SOURCES,
    base_data=DATA_FRAME,
    comparison_data=DATA_FRAME,
)
def test_comparison_to_base_example(multi_source_batch: MultiSourceBatch):
    # placeholder test to demo fixture
    base_data_source = multi_source_batch.base_batch.datasource
    context = base_data_source.data_context
    if context is None:
        raise ValueError("DataContext cannot be None")
    comparison_data_source = context.data_sources.get(
        multi_source_batch.comparison_data_source_name
    )
    assert base_data_source != comparison_data_source



from __future__ import annotations

import logging
from itertools import chain
from typing import TYPE_CHECKING, List, Optional, Sequence

from great_expectations.compatibility.typing_extensions import override
from great_expectations.experimental.metric_repository.metric_retriever import (
    MetricRetriever,
)
from great_expectations.experimental.metric_repository.metrics import (
    ColumnMetric,
    Metric,
    MetricTypes,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.datasource.fluent.batch_request import BatchRequest
    from great_expectations.validator.validator import (
        Validator,
    )


class MetricListMetricRetriever(MetricRetriever):
    def __init__(self, context: AbstractDataContext):
        super().__init__(context=context)
        self._validator: Validator | None = None

    @override
    def get_metrics(
        self,
        batch_request: BatchRequest,
        metric_list: Optional[List[MetricTypes]] = None,
    ) -> Sequence[Metric]:
        metrics_result: List[Metric] = []

        if not metric_list:
            raise ValueError("metric_list cannot be empty")  # noqa: TRY003 # FIXME CoP

        self._check_valid_metric_types(metric_list)

        table_metrics = self._calculate_table_metrics(
            batch_request=batch_request, metric_list=metric_list
        )
        metrics_result.extend(table_metrics)

        # exit early if only Table Metrics exist
        if not self._column_metrics_in_metric_list(metric_list):
            # if no column metrics are present in the metric list, we can return the table metrics
            return metrics_result

        if MetricTypes.TABLE_COLUMN_TYPES not in metric_list:
            logger.warning(
                "TABLE_COLUMN_TYPES metric is required to compute column metrics. \
                Skipping column metrics."
            )
            return metrics_result

        table_column_types = list(
            filter(lambda m: m.metric_name == MetricTypes.TABLE_COLUMN_TYPES, table_metrics)
        )[0]

        # We need to skip columns that do not report a type, because the metric computation
        # to determine semantic type will fail.
        exclude_column_names = self._get_columns_to_exclude(table_column_types)

        numeric_column_names = self._get_numeric_column_names(
            batch_request=batch_request, exclude_column_names=exclude_column_names
        )
        timestamp_column_names = self._get_timestamp_column_names(
            batch_request=batch_request, exclude_column_names=exclude_column_names
        )
        numeric_column_metrics = self._get_numeric_column_metrics(
            metric_list, batch_request, numeric_column_names
        )
        timestamp_column_metrics = self._get_timestamp_column_metrics(
            metric_list, batch_request, timestamp_column_names
        )
        all_column_names: List[str] = self._get_all_column_names(table_metrics)
        non_numeric_column_metrics = self._get_non_numeric_column_metrics(
            metric_list, batch_request, all_column_names
        )

        bundled_list = list(
            chain(
                table_metrics,
                numeric_column_metrics,
                timestamp_column_metrics,
                non_numeric_column_metrics,
            )
        )

        return bundled_list

    def _get_non_numeric_column_metrics(
        self,
        metrics_list: List[MetricTypes],
        batch_request: BatchRequest,
        column_list: List[str],
    ) -> Sequence[Metric]:
        """Calculate column metrics for non-numeric columns.

        Args:
            metrics_list (List[MetricTypes]): list of metrics sent from Agent.
            batch_request (BatchRequest): for current batch.
            column_list (List[str]): list of non-numeric columns.

        Returns:
            Sequence[Metric]: List of metrics for non-numeric columns.
        """
        # currently only the null-count is supported. If more metrics are added, this set will need to be updated.  # noqa: E501 # FIXME CoP
        column_metric_names = {MetricTypes.COLUMN_NULL_COUNT}
        metrics: list[Metric] = []
        metrics_list_as_set = set(metrics_list)
        metrics_to_calculate = sorted(column_metric_names.intersection(metrics_list_as_set))

        if not metrics_to_calculate:
            return metrics
        else:
            return self._get_column_metrics(
                batch_request=batch_request,
                column_list=column_list,
                column_metric_names=list(metrics_to_calculate),
                column_metric_type=ColumnMetric[int],
            )

    def _get_numeric_column_metrics(
        self,
        metrics_list: List[MetricTypes],
        batch_request: BatchRequest,
        column_list: List[str],
    ) -> Sequence[Metric]:
        """Calculate column metrics for numeric columns.

        Args:
            metrics_list (List[MetricTypes]): list of metrics sent from Agent.
            batch_request (BatchRequest): for current batch.
            column_list (List[str]): list of numeric columns.

        Returns:
            Sequence[Metric]: List of metrics for numeric columns.
        """
        metrics: list[Metric] = []
        column_metric_names = {
            MetricTypes.COLUMN_MIN,
            MetricTypes.COLUMN_MAX,
            MetricTypes.COLUMN_MEAN,
            MetricTypes.COLUMN_MEDIAN,
        }
        metrics_list_as_set = set(metrics_list)
        metrics_to_calculate = sorted(column_metric_names.intersection(metrics_list_as_set))
        if not metrics_to_calculate:
            return metrics

        return self._get_column_metrics(
            batch_request=batch_request,
            column_list=column_list,
            column_metric_names=list(metrics_to_calculate),
            column_metric_type=ColumnMetric[float],
        )

    def _get_timestamp_column_metrics(
        self,
        metrics_list: List[MetricTypes],
        batch_request: BatchRequest,
        column_list: List[str],
    ) -> Sequence[Metric]:
        """Calculate column metrics for timestamp columns.

        Args:
            metrics_list (List[MetricTypes]): list of metrics sent from Agent.
            batch_request (BatchRequest): for current batch.
            column_list (List[str]): list of timestamp columns.

        Returns:
            Sequence[Metric]: List of metrics for timestamp columns.
        """
        metrics: list[Metric] = []
        column_metric_names = {
            MetricTypes.COLUMN_MIN,
            MetricTypes.COLUMN_MAX,
            # MetricTypes.COLUMN_MEAN,  # Currently not supported for timestamp in Snowflake
            #  MetricTypes.COLUMN_MEDIAN,  # Currently not supported for timestamp in Snowflake
        }
        metrics_list_as_set = set(metrics_list)
        metrics_to_calculate = sorted(column_metric_names.intersection(metrics_list_as_set))
        if not metrics_to_calculate:
            return metrics

        # Note: Timestamps are returned as strings for Snowflake, this may need to be adjusted
        # when we support other datasources. For example in Pandas, timestamps can be returned as Timestamp().  # noqa: E501 # FIXME CoP
        return self._get_column_metrics(
            batch_request=batch_request,
            column_list=column_list,
            column_metric_names=list(metrics_to_calculate),
            column_metric_type=ColumnMetric[str],
        )

    def _calculate_table_metrics(
        self, batch_request: BatchRequest, metric_list: List[MetricTypes]
    ) -> List[Metric]:
        """Calculate table metrics, which include row_count, column names and types.

        Args:
            metrics_list (List[MetricTypes]): list of metrics sent from Agent.
            batch_request (BatchRequest): for current batch.

        Returns:
            Sequence[Metric]: List of table metrics.
        """
        metrics: List[Metric] = []
        if MetricTypes.TABLE_ROW_COUNT in metric_list:
            metrics.append(self._get_table_row_count(batch_request=batch_request))
        if MetricTypes.TABLE_COLUMNS in metric_list:
            metrics.append(self._get_table_columns(batch_request=batch_request))
        if MetricTypes.TABLE_COLUMN_TYPES in metric_list:
            metrics.append(self._get_table_column_types(batch_request=batch_request))
        return metrics

    def _check_valid_metric_types(self, metric_list: List[MetricTypes]) -> bool:
        """Check whether all the metric types in the list are valid.

        Args:
            metric_list (List[MetricTypes]): list of MetricTypes that are passed in to MetricListMetricRetriever.

        Returns:
            bool: True if all the metric types in the list are valid, False otherwise.
        """  # noqa: E501 # FIXME CoP
        return all(metric in MetricTypes for metric in metric_list)

    def _column_metrics_in_metric_list(self, metric_list: List[MetricTypes]) -> bool:
        """Helper method to check whether any column metrics are present in the metric list.

        Args:
            metric_list (List[MetricTypes]): list of MetricTypes that are passed in to MetricListMetricRetriever.

        Returns:
            bool: True if any column metrics are present in the metric list, False otherwise.
        """  # noqa: E501 # FIXME CoP
        column_metrics: List[MetricTypes] = [
            MetricTypes.COLUMN_MIN,
            MetricTypes.COLUMN_MAX,
            MetricTypes.COLUMN_MEDIAN,
            MetricTypes.COLUMN_MEAN,
            MetricTypes.COLUMN_NULL_COUNT,
        ]
        return any(metric in metric_list for metric in column_metrics)



import pytest


@pytest.mark.parametrize(
    "context_fixture_name",
    [
        pytest.param("ephemeral_context_with_defaults", marks=[pytest.mark.unit]),
        pytest.param("empty_data_context", marks=[pytest.mark.filesystem]),
        pytest.param("empty_cloud_context_fluent", marks=[pytest.mark.cloud]),
    ],
)
def test_fluent_datasources_show_when_printed(
    context_fixture_name: str,
    request,
) -> None:
    """Verify that fluent datasources show up when printed.
    This ends up testing both add and update (since adding assets and batch defs
    are implemented as updates on the datasource)
    """
    context = request.getfixturevalue(context_fixture_name)

    ds_name = "my-datasource"
    asset_name = "my-asset"
    batch_definition_name = "my-batch_definition"
    (
        context.data_sources.add_pandas(name=ds_name)
        .add_dataframe_asset(name=asset_name)
        .add_batch_definition_whole_dataframe(name=batch_definition_name)
    )

    str_output = context.__str__()
    repr_output = context.__repr__()

    for output in (str_output, repr_output):
        assert ds_name in output
        assert asset_name in output
        assert batch_definition_name in output


@pytest.mark.parametrize(
    "context_fixture_name",
    [
        pytest.param("ephemeral_context_with_defaults", marks=[pytest.mark.unit]),
        pytest.param("empty_data_context", marks=[pytest.mark.filesystem]),
        pytest.param("empty_cloud_context_fluent", marks=[pytest.mark.cloud]),
    ],
)
def test_deleted_fluent_datasources_do_not_show_when_printed(
    context_fixture_name: str,
    request,
) -> None:
    context = request.getfixturevalue(context_fixture_name)

    ds_name = "my-datasource"
    asset_name = "my-asset"
    batch_definition_name = "my-batch_definition"
    (
        context.data_sources.add_pandas(name=ds_name)
        .add_dataframe_asset(name=asset_name)
        .add_batch_definition_whole_dataframe(name=batch_definition_name)
    )

    context.delete_datasource(ds_name)

    str_output = context.__str__()
    repr_output = context.__repr__()

    for output in (str_output, repr_output):
        assert ds_name not in output
        assert asset_name not in output
        assert batch_definition_name not in output



from __future__ import annotations

from typing import List

import pytest

from great_expectations.datasource.fluent.data_connector.batch_filter import (
    BatchFilter,
    build_batch_filter,
)


@pytest.mark.unit
@pytest.mark.parametrize(
    "data_connector_query_dict, parsed_batch_slice, sliced_list",
    [
        pytest.param(
            {
                "index": "[4:9:2]",
            },
            slice(4, 9, 2),
            [4, 6, 8],
            id="batch_slice: str (with square brackets); (start, stop, step)",
        ),
        pytest.param(
            {
                "index": "4:9:2",
            },
            slice(4, 9, 2),
            [4, 6, 8],
            id="batch_slice: str (without square brackets); (start, stop, step)",
        ),
        pytest.param(
            {
                "index": "3:",
            },
            slice(3, None, None),
            [3, 4, 5, 6, 7, 8, 9],
            id="batch_slice: str (without square brackets, forward traversal at start); (start, stop=None, step=None)",  # noqa: E501 # FIXME CoP
        ),
        pytest.param(
            {
                "index": ":3",
            },
            slice(None, 3, None),
            [0, 1, 2],
            id="batch_slice: str (without square brackets); (start=None, stop, step=None)",
        ),
        pytest.param(
            {
                "index": "[1:4]",
            },
            slice(1, 4, None),
            [1, 2, 3],
            id="batch_slice: str (with square brackets); (start, stop, step=None)",
        ),
        pytest.param(
            {
                "index": "[-5:]",
            },
            slice(-5, None, None),
            [5, 6, 7, 8, 9],
            id="batch_slice: str (with square brackets); (start, stop=None, step=None)",
        ),
        pytest.param(
            {
                "index": (1, 7, 3),
            },
            slice(1, 7, 3),
            [1, 4],
            id="batch_slice: tuple; (start, stop, step)",
        ),
        pytest.param(
            {
                "index": 0,
            },
            slice(0, 1, None),
            [0],
            id="batch_slice: zero int; (start, stop=None, step=None)",
        ),
        pytest.param(
            {
                "index": -1,
            },
            slice(-1, None, None),
            [9],
            id="batch_slice: negative int; (start, stop=None, step=None)",
        ),
        pytest.param(
            {
                "index": "0",
            },
            slice(0, 1, None),
            [0],
            id="batch_slice: str (zero int); (start, stop=None, step=None)",
        ),
        pytest.param(
            {
                "index": "-1",
            },
            slice(-1, None, None),
            [9],
            id="batch_slice: str (negative int); (start, stop=None, step=None)",
        ),
        pytest.param(
            {
                "index": slice(-1, 0, -1),
            },
            slice(-1, 0, -1),
            [9, 8, 7, 6, 5, 4, 3, 2, 1],
            id="batch_slice: slice (reverse traversal); (negative_start, zero_stop, negative_step)",
        ),
        pytest.param(
            {
                "index": "::",
            },
            slice(None, None, None),
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            id="batch_slice: str (full forward traversal); (start=None, stop=None, step=1)",
        ),
        pytest.param(
            {
                "index": "::2",
            },
            slice(None, None, 2),
            [0, 2, 4, 6, 8],
            id="batch_slice: str (full forward traversal with step=2); (start=None, stop=None, step=2)",  # noqa: E501 # FIXME CoP
        ),
        pytest.param(
            {
                "index": "::-1",
            },
            slice(None, None, -1),
            [9, 8, 7, 6, 5, 4, 3, 2, 1, 0],
            id="batch_slice: str (full reverse traversal); (start=None, stop=None, step=-1)",
        ),
    ],
)
def test_batch_filter_parse_batch_slice(
    data_connector_query_dict: dict,
    parsed_batch_slice: slice,
    sliced_list: List[int],
):
    original_list: List[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    batch_filter_obj: BatchFilter = build_batch_filter(
        data_connector_query_dict=data_connector_query_dict  # type: ignore[arg-type] # FIXME CoP
    )
    assert batch_filter_obj.index == parsed_batch_slice
    assert original_list[parsed_batch_slice] == sliced_list



from __future__ import annotations

import pathlib
import re
import uuid
from typing import TYPE_CHECKING, Optional

import pytest

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.partitioners import FileNamePartitionerYearly
from great_expectations.datasource.fluent.batch_request import BatchParameters
from great_expectations.datasource.fluent.interfaces import Batch, DataAsset
from great_expectations.exceptions import (
    BatchDefinitionNotAddedError,
    BatchDefinitionNotFreshError,
    ResourceFreshnessAggregateError,
)
from great_expectations.exceptions.exceptions import (
    BatchDefinitionNotFoundError,
    DataAssetNotFoundError,
    DatasourceNotFoundError,
)

if TYPE_CHECKING:
    from typing import List

    import pytest_mock

    from great_expectations.data_context.data_context.ephemeral_data_context import (
        EphemeralDataContext,
    )
    from great_expectations.datasource.fluent.batch_request import BatchRequest


class DataAssetForTests(DataAsset):
    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
        raise NotImplementedError

    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch:
        raise NotImplementedError


@pytest.fixture
def mock_data_asset(monkeypatch, mocker: pytest_mock.MockerFixture) -> DataAsset:
    monkeypatch.setattr(DataAsset, "build_batch_request", mocker.Mock())
    data_asset: DataAsset = DataAssetForTests(name="my_data_asset", type="table")

    return data_asset


@pytest.mark.parametrize(
    "batch_parameters",
    [
        (None,),
        ({"foo": "bar"},),
    ],
)
@pytest.mark.unit
def test_build_batch_request(
    batch_parameters: Optional[BatchParameters],
    mock_data_asset: DataAsset,
    mocker: pytest_mock.MockerFixture,
):
    batching_regex = re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2}).csv")
    partitioner = FileNamePartitionerYearly(regex=batching_regex)
    batch_definition = BatchDefinition(
        name="test_batch_definition",
        partitioner=partitioner,
    )
    batch_definition.set_data_asset(mock_data_asset)

    batch_definition.build_batch_request(batch_parameters=batch_parameters)

    mock_build_batch_request = batch_definition.data_asset.build_batch_request
    assert isinstance(mock_build_batch_request, mocker.Mock)
    mock_build_batch_request.assert_called_once_with(
        options=batch_parameters,
        partitioner=partitioner,
    )


@pytest.mark.unit
def test_get_batch_retrieves_only_batch(mocker: pytest_mock.MockFixture):
    # Arrange
    batch_definition = BatchDefinition[None](name="test_batch_definition")
    mock_asset = mocker.Mock(spec=DataAsset)
    batch_definition.set_data_asset(mock_asset)

    mock_batch = mocker.Mock(spec=Batch)
    mock_asset.get_batch.return_value = mock_batch

    # Act
    batch = batch_definition.get_batch()

    # Assert
    assert batch == mock_batch
    mock_asset.get_batch.assert_called_once_with(batch_definition.build_batch_request())


@pytest.mark.unit
def test_get_batch_identifiers_list(mocker: pytest_mock.MockFixture):
    # Arrange
    batch_definition = BatchDefinition[None](name="test_batch_definition")
    mock_asset = mocker.Mock(spec=DataAsset)
    batch_definition.set_data_asset(mock_asset)

    mock_batch_identifiers_list = [{"foo": "bar"}, {"baz": "qux"}]
    mock_asset.get_batch_identifiers_list.return_value = mock_batch_identifiers_list

    # Act
    batch_identifiers_list = batch_definition.get_batch_identifiers_list()

    # Assert
    assert batch_identifiers_list == mock_batch_identifiers_list
    mock_asset.get_batch_identifiers_list.assert_called_once_with(
        batch_definition.build_batch_request()
    )


@pytest.mark.unit
def test_get_batch_identifiers_list_with_batch_parameters(mocker: pytest_mock.MockFixture):
    # Arrange
    batch_definition = BatchDefinition[None](name="test_batch_definition")
    mock_asset = mocker.Mock(spec=DataAsset)
    batch_definition.set_data_asset(mock_asset)

    mock_batch_identifiers_list = [{"foo": "bar"}, {"baz": "qux"}]
    mock_asset.get_batch_identifiers_list.return_value = mock_batch_identifiers_list

    # Act
    batch_parameters: BatchParameters = {"path": "my_path"}
    batch_identifiers_list = batch_definition.get_batch_identifiers_list(batch_parameters)

    # Assert
    assert batch_identifiers_list == mock_batch_identifiers_list
    mock_asset.get_batch_identifiers_list.assert_called_once_with(
        batch_definition.build_batch_request(batch_parameters)
    )


@pytest.mark.unit
def test_identifier_bundle_success(in_memory_runtime_context):
    context = in_memory_runtime_context
    ds = context.data_sources.add_pandas("pandas_datasource")
    asset = ds.add_csv_asset("my_asset", "data.csv")
    batch_definition = asset.add_batch_definition("my_batch_definition")

    result = batch_definition.identifier_bundle()
    assert result.datasource.name == "pandas_datasource" and result.datasource.id is not None
    assert result.asset.name == "my_asset" and result.asset.id is not None
    assert (
        result.batch_definition.name == "my_batch_definition"
        and result.batch_definition.id is not None
    )


@pytest.mark.unit
def test_identifier_bundle_no_id_raises_error(in_memory_runtime_context):
    context = in_memory_runtime_context
    ds = context.data_sources.add_pandas("pandas_datasource")
    asset = ds.add_csv_asset("my_asset", "data.csv")
    batch_definition = asset.add_batch_definition("my_batch_definition")

    batch_definition.id = None

    with pytest.raises(ResourceFreshnessAggregateError) as e:
        batch_definition.identifier_bundle()

    assert len(e.value.errors) == 1
    assert isinstance(e.value.errors[0], BatchDefinitionNotAddedError)


@pytest.mark.parametrize(
    "id,is_fresh,num_errors",
    [
        pytest.param(str(uuid.uuid4()), True, 0, id="added"),
        pytest.param(None, False, 1, id="not_added"),
    ],
)
@pytest.mark.unit
def test_is_fresh_is_added(
    in_memory_runtime_context, id: str | None, is_fresh: bool, num_errors: int
):
    context = in_memory_runtime_context
    batch_definition = (
        context.data_sources.add_pandas(name="my_pandas_ds")
        .add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
        .add_batch_definition(name="my_batch_def")
    )
    batch_definition.id = id  # Fluent API will add an ID but manually overriding for test
    diagnostics = batch_definition.is_fresh()

    assert diagnostics.success is is_fresh
    assert len(diagnostics.errors) == num_errors
    assert all(isinstance(err, BatchDefinitionNotAddedError) for err in diagnostics.errors)


@pytest.mark.cloud
def test_is_fresh_freshness(empty_cloud_context_fluent):
    # Ephemeral/file use a cacheable datasource dict so freshness
    # with batch definitions is a Cloud-only concern
    context = empty_cloud_context_fluent
    batch_definition = (
        context.data_sources.add_pandas(name="my_pandas_ds")
        .add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
        .add_batch_definition(name="my_batch_def")
    )

    batching_regex = re.compile(r"data_(?P<year>\d{4})-(?P<month>\d{2}).csv")
    partitioner = FileNamePartitionerYearly(regex=batching_regex)
    batch_definition.partitioner = partitioner

    diagnostics = batch_definition.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], BatchDefinitionNotFreshError)


@pytest.mark.unit
def test_is_fresh_fails_on_datasource_retrieval(in_memory_runtime_context):
    context = in_memory_runtime_context
    datasource = context.data_sources.add_pandas(name="my_pandas_ds")
    asset = datasource.add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
    batch_definition = asset.add_batch_definition(name="my_batch_def")

    context.delete_datasource("my_pandas_ds")

    diagnostics = batch_definition.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], DatasourceNotFoundError)


@pytest.mark.unit
def test_is_fresh_fails_on_asset_retrieval(in_memory_runtime_context):
    context = in_memory_runtime_context
    datasource = context.data_sources.add_pandas(name="my_pandas_ds")
    asset = datasource.add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
    batch_definition = asset.add_batch_definition(name="my_batch_def")

    datasource.delete_asset("my_csv_asset")

    diagnostics = batch_definition.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], DataAssetNotFoundError)


@pytest.mark.unit
def test_is_fresh_fails_on_batch_definition_retrieval(in_memory_runtime_context):
    context = in_memory_runtime_context
    datasource = context.data_sources.add_pandas(name="my_pandas_ds")
    asset = datasource.add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
    batch_definition = asset.add_batch_definition(name="my_batch_def")

    asset.delete_batch_definition("my_batch_def")

    diagnostics = batch_definition.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], BatchDefinitionNotFoundError)


@pytest.mark.unit
def test_save(in_memory_runtime_context: EphemeralDataContext):
    context = in_memory_runtime_context

    ds_name = "my_pandas_ds"
    asset_name = "my_csv_asset"
    batch_def_name = "my_batch_def"

    datasource = context.data_sources.add_pandas(name=ds_name)
    asset = datasource.add_csv_asset(name=asset_name, filepath_or_buffer=pathlib.Path("data.csv"))
    batch_definition = asset.add_batch_definition(name=batch_def_name)

    assert batch_definition.partitioner is None

    batch_definition.partitioner = FileNamePartitionerYearly(regex=re.compile("my_regex"))
    batch_definition.save()

    retrieved_datasource = context.data_sources.get(name=ds_name)
    retrieved_asset = retrieved_datasource.get_asset(name=asset_name)
    retrieved_batch_definition = retrieved_asset.get_batch_definition(name=batch_def_name)

    assert retrieved_batch_definition.partitioner



from typing import ClassVar

from great_expectations.expectations.metrics.query_metrics.query_table.query_table import QueryTable


class UnexpectedRowsQueryTable(QueryTable):
    metric_name = "unexpected_rows_query.table"
    value_keys = ("unexpected_rows_query",)

    query_param_name: ClassVar[str] = "unexpected_rows_query"



from __future__ import annotations

import operator
from functools import reduce
from typing import List, Tuple

# The following method is used to ensure that the dictionary used to verify universal configuration elements
# remains the same in all the configuration tests.  Users may disregard it.


def _get_items_by_path(root_dictionary: dict, keys: Tuple[str]) -> Tuple:
    try:
        return "/".join(keys), reduce(operator.getitem, keys, root_dictionary)
    except Exception:
        print(root_dictionary, keys)


def _gather_key_paths_from_dict(
    target_dict: dict, current_path: List[str] | None = None
) -> Tuple[List[Tuple[str]], List[str]]:
    key_paths: List[Tuple[str]] = []
    full_paths: List[Tuple[str]] = []
    for key, value in target_dict.items():
        if isinstance(value, dict):
            if current_path:
                next_path = current_path[:]
                full_paths.append(tuple(next_path))
                next_path.append(key)
                full_paths.append(tuple(next_path))
            else:
                next_path = [key]
            if value:
                new_key_paths, new_full_paths = _gather_key_paths_from_dict(value, next_path)
                key_paths.extend(new_key_paths)
                full_paths.extend(new_full_paths)
            else:
                # If this is an empty dictionary, then there will be no further nested keys to gather.
                key_paths.append(tuple(next_path))
                full_paths.append(tuple(next_path))
        elif current_path:
            next_path = current_path[:]
            next_path.append(key)
            key_paths.append(tuple(next_path))
            full_paths.append(tuple(next_path))
        else:
            full_paths.append((key,))
            key_paths.append((key,))
    return key_paths, full_paths


def is_subset(subset, superset, test_mode=True):
    superset_key_paths, superset_full_paths = _gather_key_paths_from_dict(superset)
    superset_full_paths = ["/".join(x) for x in superset_full_paths]
    # print(superset_key_paths)
    # print(superset_full_paths)
    # print("")

    subset_key_paths, subset_full_paths = _gather_key_paths_from_dict(subset)
    subset_full_paths = ["/".join(x) for x in subset_full_paths]
    # print(subset_key_paths)
    # print(subset_full_paths)
    # print("")

    subset_items = [_get_items_by_path(subset, key_path) for key_path in subset_key_paths]
    superset_items = [_get_items_by_path(superset, key_path) for key_path in superset_key_paths]
    # key_paths that point to empty values should only be checked to see if they exist as key_paths in the superset
    # empty_paths = [x[0] for x in subset_items if not x[1]]
    subset_items = [x for x in subset_items if x[1]]
    #
    # superset_items = [_get_items_by_path(superset, key_path) for key_path in key_paths]

    # Test that populated subset items correspond to items in the superset
    if subset_items:
        items_test = all(item in superset_items for item in subset_items)
    else:
        items_test = True

    # Test that the subset key paths correspond to key paths in the superset
    keys_test = all(item in superset_full_paths for item in subset_full_paths)

    if test_mode is True:
        assert all([items_test, keys_test]), (
            f"\nEITHER:\n{subset_items}\n is not a subset of \n{superset_items}\n\nOR:\n\n{subset_full_paths}\n is not a subset of \n{superset_full_paths}"
        )
    else:
        assert all([items_test, keys_test]) is False, (
            f"\nBOTH:\n{subset_items} is a subset of \n{superset_items}\nAND:\n{subset_full_paths} is a subset of \n{superset_full_paths}"
        )


if __name__ == "__main__":
    superset = {
        "test": "test",
        "test2": {"test3": "test3", "test4": {"test5": "test5"}},
    }

    # test empty top level subset:
    subset = {"test": "", "test2": {}}
    is_subset(subset, superset)

    # test correct top level subset:
    subset = {"test": "test", "test2": {}}
    is_subset(subset, superset)

    # test incorrect top level subset:
    subset = {"test": "test", "incorrect": {}}
    is_subset(subset, superset, False)

    # test incorrect top level subset:
    subset = {"test": "incorrect", "test2": {}}
    is_subset(subset, superset, False)

    ## Nested Tests:
    # test empty nested level 1 subsets:
    subset = {"test": "test", "test2": {"test3": "", "test4": {"test5": "test5"}}}
    is_subset(subset, superset)

    subset = {"test": "test", "test2": {"test3": "test3", "test4": {}}}
    is_subset(subset, superset)

    # test correct nested level 1 subsets:
    subset = {"test": "test", "test2": {"test3": "test3", "test4": {"test5": "test5"}}}
    is_subset(subset, superset)

    # test incorrect nested level 1 subsets:
    subset = {
        "test": "test",
        "test2": {"test3": "incorrect", "test4": {"test5": "test5"}},
    }
    is_subset(subset, superset, False)

    subset = {
        "test": "test",
        "test2": {"test3": "test3", "test4": {"incorrect": "test5"}},
    }
    is_subset(subset, superset, False)

    subset = {
        "test": "test",
        "test2": {"test3": "test3", "test4": {"test5": "incorrect"}},
    }
    is_subset(subset, superset, False)



from __future__ import annotations

import hashlib
import json
from typing import Any, Set, TypeVar, Union

from great_expectations.compatibility.typing_extensions import override
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP

T = TypeVar("T")


IDDictID = Union[str, tuple[()]]


class IDDict(dict):
    _id_ignore_keys: Set[str] = set()

    def to_id(self, id_keys=None, id_ignore_keys=None) -> IDDictID:
        if id_keys is None:
            id_keys = self.keys()
        if id_ignore_keys is None:
            id_ignore_keys = self._id_ignore_keys
        id_keys = set(id_keys) - set(id_ignore_keys)
        if len(id_keys) == 0:
            return ()
        elif len(id_keys) == 1:
            key = list(id_keys)[0]
            return f"{key}={self[key]!s}"

        _id_dict = convert_to_json_serializable(data={k: self[k] for k in id_keys})
        return hashlib.md5(json.dumps(_id_dict, sort_keys=True).encode("utf-8")).hexdigest()

    @override
    def __hash__(self) -> int:  # type: ignore[override] # FIXME CoP
        """Overrides the default implementation"""
        _result_hash: int = hash(self.to_id())
        return _result_hash


def deep_convert_properties_iterable_to_id_dict(
    source: Union[T, dict],
) -> Union[T, IDDict]:
    if isinstance(source, dict):
        return _deep_convert_properties_iterable_to_id_dict(source=IDDict(source))

    # Must allow for non-dictionary source types, since their internal nested structures may contain dictionaries.  # noqa: E501 # FIXME CoP
    if isinstance(source, (list, set, tuple)):
        data_type: type = type(source)

        element: Any
        return data_type(
            [deep_convert_properties_iterable_to_id_dict(source=element) for element in source]
        )

    return source


def _deep_convert_properties_iterable_to_id_dict(source: dict) -> IDDict:
    key: str
    value: Any
    for key, value in source.items():
        if isinstance(value, dict):
            source[key] = _deep_convert_properties_iterable_to_id_dict(source=value)
        elif isinstance(value, (list, set, tuple)):
            data_type: type = type(value)

            element: Any
            source[key] = data_type(
                [deep_convert_properties_iterable_to_id_dict(source=element) for element in value]
            )

    return IDDict(source)


class BatchKwargs(IDDict):
    pass


class BatchSpec(IDDict):
    pass


class MetricKwargs(IDDict):
    pass



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, List, Optional

from great_expectations.experimental.metric_repository.metrics import (
    Metric,
    MetricRun,
    MetricTypes,
)

if TYPE_CHECKING:
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.datasource.fluent.interfaces import BatchRequest
    from great_expectations.experimental.metric_repository.metric_retriever import (
        MetricRetriever,
    )


class BatchInspector:
    """A BatchInspector is responsible for computing metrics for a batch of data.

    It uses MetricRetriever objects to retrieve metrics.
    """

    def __init__(self, context: AbstractDataContext, metric_retrievers: list[MetricRetriever]):
        self._context = context
        self._metric_retrievers = metric_retrievers

    def compute_metric_list_run(
        self,
        data_asset_id: uuid.UUID,
        batch_request: BatchRequest,
        metric_list: Optional[List[MetricTypes]],
    ) -> MetricRun:
        """Method that computes a MetricRun for a list of metrics.
        Called by GX Agent to compute a MetricRun as part of a RunMetricsEvent.
        Args:
            data_asset_id (uuid.UUID): current data asset id.
            batch_request (BatchRequest): BatchRequest for current batch.
            metrics_list (Optional[List[MetricTypes]]): List of metrics to compute.
        Returns:
            MetricRun: _description_
        """
        # TODO: eventually we will keep this and retire `compute_metric_run`.
        metrics: list[Metric] = []
        for metric_retriever in self._metric_retrievers:
            metrics.extend(
                metric_retriever.get_metrics(batch_request=batch_request, metric_list=metric_list)
            )
        return MetricRun(data_asset_id=data_asset_id, metrics=metrics)

    def compute_metric_run(
        self, data_asset_id: uuid.UUID, batch_request: BatchRequest
    ) -> MetricRun:
        metrics: list[Metric] = []
        for metric_retriever in self._metric_retrievers:
            metrics.extend(metric_retriever.get_metrics(batch_request=batch_request))

        return MetricRun(data_asset_id=data_asset_id, metrics=metrics)

    def _generate_run_id(self) -> uuid.UUID:
        return uuid.uuid4()



from __future__ import annotations

import datetime
import os
from typing import List

import pandas as pd
import pytest
from dateutil.parser import parse

from great_expectations.compatibility.sqlalchemy_compatibility_wrappers import (
    add_dataframe_to_db,
)
from great_expectations.core.batch_spec import SqlAlchemyDatasourceBatchSpec
from great_expectations.core.id_dict import BatchSpec
from great_expectations.data_context.util import file_relative_path
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from great_expectations.execution_engine.partition_and_sample.sqlalchemy_data_sampler import (
    SqlAlchemyDataSampler,
)
from great_expectations.execution_engine.sqlalchemy_batch_data import (
    SqlAlchemyBatchData,
)
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect
from great_expectations.self_check.util import build_sa_execution_engine
from great_expectations.util import import_library_module

try:
    sqlalchemy = pytest.importorskip("sqlalchemy")
except ImportError:
    sqlalchemy = None


pytestmark = [
    pytest.mark.sqlalchemy_version_compatibility,
    pytest.mark.external_sqldialect,
]


@pytest.mark.unit
@pytest.mark.parametrize(
    "underscore_prefix",
    [
        pytest.param("_", id="underscore prefix"),
        pytest.param("", id="no underscore prefix"),
    ],
)
@pytest.mark.parametrize(
    "sampler_method_name",
    [
        pytest.param(sampler_method_name, id=sampler_method_name)
        for sampler_method_name in [
            "sample_using_limit",
            "sample_using_random",
            "sample_using_mod",
            "sample_using_a_list",
            "sample_using_md5",
        ]
    ],
)
def test_get_sampler_method(underscore_prefix: str, sampler_method_name: str):
    """What does this test and why?

    This test is to ensure that the sampler methods are accessible with and without underscores.
    When new sampling methods are added, the parameter list should be updated.
    """
    data_partitioner: SqlAlchemyDataSampler = SqlAlchemyDataSampler()

    sampler_method_name_with_prefix = f"{underscore_prefix}{sampler_method_name}"

    assert data_partitioner.get_sampler_method(sampler_method_name_with_prefix) == getattr(
        data_partitioner, sampler_method_name
    )


def clean_query_for_comparison(query_string: str) -> str:
    """Remove whitespace and case from query for easy comparison.

    Args:
        query_string: query string to convert.

    Returns:
        String with removed whitespace and converted to lowercase.
    """
    """Remove """
    return query_string.replace("\n", "").replace("\t", "").replace(" ", "").lower()


@pytest.fixture
def dialect_name_to_sql_statement():
    def _dialect_name_to_sql_statement(dialect_name: GXSqlDialect) -> str:
        dialect_name_to_sql_statement: dict = {
            GXSqlDialect.POSTGRESQL: "SELECT * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE TRUE LIMIT 10",  # noqa: E501 # FIXME CoP
            GXSqlDialect.MYSQL: "SELECT * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE TRUE = 1 LIMIT 10",
            GXSqlDialect.ORACLE: "SELECT * FROM test_schema_name.test_table WHERE 1 = 1 AND ROWNUM <= 10",  # noqa: E501 # FIXME CoP
            GXSqlDialect.MSSQL: "SELECT TOP 10 * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE 1 = 1",
            GXSqlDialect.SQLITE: "SELECT * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE 1 = 1 LIMIT 10 OFFSET 0",  # noqa: E501 # FIXME CoP
            GXSqlDialect.BIGQUERY: "SELECT * FROM `TEST_SCHEMA_NAME`.`TEST_TABLE` WHERE TRUE LIMIT 10",  # noqa: E501 # FIXME CoP
            GXSqlDialect.SNOWFLAKE: "SELECT * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE TRUE LIMIT 10",
            GXSqlDialect.REDSHIFT: "SELECT * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE TRUE LIMIT 10",
            GXSqlDialect.AWSATHENA: 'SELECT * FROM "TEST_SCHEMA_NAME"."TEST_TABLE" WHERE TRUE LIMIT 10',  # noqa: E501 # FIXME CoP
            GXSqlDialect.DREMIO: 'SELECT * FROM "TEST_SCHEMA_NAME"."TEST_TABLE" WHERE 1 = 1 LIMIT 10',  # noqa: E501 # FIXME CoP
            GXSqlDialect.TERADATASQL: "SELECT TOP 10 * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE 1 = 1",  # noqa: E501 # FIXME CoP
            GXSqlDialect.TRINO: "SELECT * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE TRUE LIMIT 10",
            GXSqlDialect.HIVE: "SELECT * FROM `TEST_SCHEMA_NAME`.`TEST_TABLE` WHERE TRUE LIMIT 10",
            GXSqlDialect.VERTICA: "SELECT * FROM TEST_SCHEMA_NAME.TEST_TABLE WHERE TRUE LIMIT 10",
        }
        return dialect_name_to_sql_statement[dialect_name]

    return _dialect_name_to_sql_statement


@pytest.fixture
def pytest_parsed_arguments(request):
    return request.config.option


# Despite being parameterized over GXSqlDialect, this test skips if the flag corresponding to that dialect isn't  # noqa: E501 # FIXME CoP
# passed in. Most of these dialects are never run in CI.
@pytest.mark.all_backends
@pytest.mark.parametrize(
    "dialect_name",
    [
        pytest.param(dialect_name, id=dialect_name.value, marks=pytest.mark.external_sqldialect)
        for dialect_name in GXSqlDialect.get_all_dialects()
    ],
)
def test_sample_using_limit_builds_correct_query_where_clause_none(  # noqa: C901 # FIXME CoP
    dialect_name: GXSqlDialect,
    dialect_name_to_sql_statement,
    sa,
    pytest_parsed_arguments,
):
    """What does this test and why?

    partition_on_limit should build the appropriate query based on input parameters.
    This tests dialects that differ from the standard dialect, not each dialect exhaustively.
    """
    if hasattr(pytest_parsed_arguments, str(dialect_name.value)):
        if not getattr(pytest_parsed_arguments, str(dialect_name.value)):
            pytest.skip(
                f"Skipping {dialect_name.value!s} since the --{dialect_name.value!s} pytest flag was not set"  # noqa: E501 # FIXME CoP
            )
    else:
        pytest.skip(
            f"Skipping {dialect_name.value!s} since the dialect is not runnable via pytest flag"
        )

    # 1. Setup
    class MockSqlAlchemyExecutionEngine:
        def __init__(self, dialect_name: GXSqlDialect):
            self._dialect_name = dialect_name
            self._connection_string = self.dialect_name_to_connection_string(dialect_name)

        DIALECT_TO_CONNECTION_STRING_STUB: dict = {
            GXSqlDialect.POSTGRESQL: "postgresql://",
            GXSqlDialect.MYSQL: "mysql+pymysql://",
            GXSqlDialect.ORACLE: "oracle+cx_oracle://",
            GXSqlDialect.MSSQL: "mssql+pyodbc://",
            GXSqlDialect.SQLITE: "sqlite:///",
            GXSqlDialect.BIGQUERY: "bigquery://",
            GXSqlDialect.SNOWFLAKE: "snowflake://",
            GXSqlDialect.REDSHIFT: "redshift+psycopg2://",
            GXSqlDialect.AWSATHENA: "awsathena+rest://@athena.us-east-1.amazonaws.com/some_test_db?s3_staging_dir=s3://some-s3-path/",
            GXSqlDialect.DREMIO: "dremio://",
            GXSqlDialect.TERADATASQL: "teradatasql://",
            GXSqlDialect.TRINO: "trino://",
            GXSqlDialect.HIVE: "hive://",
            GXSqlDialect.VERTICA: "vertica+vertica_python://",
        }

        @property
        def dialect_name(self) -> str:
            return self._dialect_name.value

        def dialect_name_to_connection_string(self, dialect_name: GXSqlDialect) -> str:
            return self.DIALECT_TO_CONNECTION_STRING_STUB.get(dialect_name)

        _BIGQUERY_MODULE_NAME = "sqlalchemy_bigquery"

        @property
        def dialect(self) -> sa.engine.Dialect:
            # TODO: AJB 20220512 move this dialect retrieval to a separate class from the SqlAlchemyExecutionEngine  # noqa: E501 # FIXME CoP
            #  and then use it here.
            dialect_name: GXSqlDialect = self._dialect_name
            if dialect_name == GXSqlDialect.ORACLE:
                # noinspection PyUnresolvedReferences
                return import_library_module(module_name="sqlalchemy.dialects.oracle").dialect()
            elif dialect_name == GXSqlDialect.SNOWFLAKE:
                # noinspection PyUnresolvedReferences
                return import_library_module(
                    module_name="snowflake.sqlalchemy.snowdialect"
                ).dialect()
            elif dialect_name == GXSqlDialect.DREMIO:
                # WARNING: Dremio Support is experimental, functionality is not fully under test
                # noinspection PyUnresolvedReferences
                return import_library_module(module_name="sqlalchemy_dremio.pyodbc").dialect()
            # NOTE: AJB 20220512 Redshift dialect is not yet fully supported.
            # The below throws an `AttributeError: type object 'RedshiftDialect_psycopg2' has no attribute 'positional'`  # noqa: E501 # FIXME CoP
            # elif dialect_name == "redshift":
            #     return import_library_module(
            #         module_name="sqlalchemy_redshift.dialect"
            #     ).RedshiftDialect
            elif dialect_name == GXSqlDialect.BIGQUERY:
                # noinspection PyUnresolvedReferences
                return import_library_module(module_name=self._BIGQUERY_MODULE_NAME).dialect()
            elif dialect_name == GXSqlDialect.TERADATASQL:
                # WARNING: Teradata Support is experimental, functionality is not fully under test
                # noinspection PyUnresolvedReferences
                return import_library_module(module_name="teradatasqlalchemy.dialect").dialect()
            else:
                return sa.create_engine(self._connection_string).dialect

    mock_execution_engine: MockSqlAlchemyExecutionEngine = MockSqlAlchemyExecutionEngine(
        dialect_name=dialect_name
    )

    data_sampler: SqlAlchemyDataSampler = SqlAlchemyDataSampler()

    # 2. Create query using sampler
    table_name: str = "test_table"
    batch_spec = BatchSpec(
        table_name=table_name,
        schema_name="test_schema_name",
        sampling_method="sample_using_limit",
        sampling_kwargs={"n": 10},
    )
    query = data_sampler.sample_using_limit(
        execution_engine=mock_execution_engine, batch_spec=batch_spec, where_clause=None
    )

    if not isinstance(query, str):
        query_str: str = clean_query_for_comparison(
            str(
                query.compile(
                    dialect=mock_execution_engine.dialect,
                    compile_kwargs={"literal_binds": True},
                )
            )
        )
    else:
        query_str: str = clean_query_for_comparison(query)

    expected: str = clean_query_for_comparison(dialect_name_to_sql_statement(dialect_name))

    assert query_str == expected


@pytest.mark.sqlite
def test_sqlite_sample_using_limit(sa):
    csv_path: str = file_relative_path(
        os.path.dirname(os.path.dirname(__file__)),  # noqa: PTH120 # FIXME CoP
        os.path.join(  # noqa: PTH118 # FIXME CoP
            "test_sets",
            "taxi_yellow_tripdata_samples",
            "ten_trips_from_each_month",
            "yellow_tripdata_sample_10_trips_from_each_month.csv",
        ),
    )
    df: pd.DataFrame = pd.read_csv(csv_path)
    engine: SqlAlchemyExecutionEngine = build_sa_execution_engine(df, sa)

    n: int = 10
    batch_spec: SqlAlchemyDatasourceBatchSpec = SqlAlchemyDatasourceBatchSpec(
        table_name="test",
        schema_name="main",
        sampling_method="sample_using_limit",
        sampling_kwargs={"n": n},
    )
    batch_data: SqlAlchemyBatchData = engine.get_batch_data(batch_spec=batch_spec)

    # Right number of rows?
    num_rows: int = batch_data.execution_engine.execute_query(
        sa.select(sa.func.count()).select_from(batch_data.selectable)
    ).scalar()
    assert num_rows == n

    # Right rows?
    rows: list[sa.RowMapping] = (
        batch_data.execution_engine.execute_query(
            sa.select(sa.text("*")).select_from(batch_data.selectable)
        )
        .mappings()
        .fetchall()
    )

    row_dates: List[datetime.datetime] = [parse(row["pickup_datetime"]) for row in rows]
    for row_date in row_dates:
        assert row_date.month == 1
        assert row_date.year == 2018


@pytest.mark.sqlite
def test_sample_using_random(sqlite_view_engine, test_df):
    my_execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(
        engine=sqlite_view_engine
    )

    p: float
    batch_spec: SqlAlchemyDatasourceBatchSpec
    batch_data: SqlAlchemyBatchData
    num_rows: int
    rows_0: List[tuple]
    rows_1: List[tuple]

    # First, make sure that degenerative case never passes.

    test_df_0: pd.DataFrame = test_df.iloc[:1]
    add_dataframe_to_db(df=test_df_0, name="test_table_0", con=my_execution_engine.engine)

    p = 1.0
    batch_spec = SqlAlchemyDatasourceBatchSpec(
        table_name="test_table_0",
        schema_name="main",
        sampling_method="_sample_using_random",
        sampling_kwargs={"p": p},
    )

    batch_data = my_execution_engine.get_batch_data(batch_spec=batch_spec)
    num_rows = batch_data.execution_engine.execute_query(
        sqlalchemy.select(sqlalchemy.func.count()).select_from(batch_data.selectable)
    ).scalar()
    assert num_rows == round(p * test_df_0.shape[0])

    rows_0: List[tuple] = batch_data.execution_engine.execute_query(
        sqlalchemy.select(sqlalchemy.text("*")).select_from(batch_data.selectable)
    ).fetchall()

    batch_data = my_execution_engine.get_batch_data(batch_spec=batch_spec)
    num_rows = batch_data.execution_engine.execute_query(
        sqlalchemy.select(sqlalchemy.func.count()).select_from(batch_data.selectable)
    ).scalar()
    assert num_rows == round(p * test_df_0.shape[0])

    rows_1: List[tuple] = batch_data.execution_engine.execute_query(
        sqlalchemy.select(sqlalchemy.text("*")).select_from(batch_data.selectable)
    ).fetchall()

    assert len(rows_0) == len(rows_1) == 1

    assert rows_0 == rows_1

    # Second, verify that realistic case always returns different random sample of rows.

    test_df_1: pd.DataFrame = test_df
    add_dataframe_to_db(df=test_df_1, name="test_table_1", con=my_execution_engine.engine)

    p = 2.0e-1
    batch_spec = SqlAlchemyDatasourceBatchSpec(
        table_name="test_table_1",
        schema_name="main",
        sampling_method="_sample_using_random",
        sampling_kwargs={"p": p},
    )

    batch_data = my_execution_engine.get_batch_data(batch_spec=batch_spec)
    num_rows = batch_data.execution_engine.execute_query(
        sqlalchemy.select(sqlalchemy.func.count()).select_from(batch_data.selectable)
    ).scalar()
    assert num_rows == round(p * test_df_1.shape[0])

    rows_0 = batch_data.execution_engine.execute_query(
        sqlalchemy.select(sqlalchemy.text("*")).select_from(batch_data.selectable)
    ).fetchall()

    batch_data = my_execution_engine.get_batch_data(batch_spec=batch_spec)
    num_rows = batch_data.execution_engine.execute_query(
        sqlalchemy.select(sqlalchemy.func.count()).select_from(batch_data.selectable)
    ).scalar()
    assert num_rows == round(p * test_df_1.shape[0])

    rows_1 = batch_data.execution_engine.execute_query(
        sqlalchemy.select(sqlalchemy.text("*")).select_from(batch_data.selectable)
    ).fetchall()

    assert len(rows_0) == len(rows_1)

    assert rows_0 != rows_1


@pytest.mark.unit
def test_sample_using_random_batch_spec_test_table_name_required():
    fake_execution_engine = None
    batch_spec = BatchSpec()
    with pytest.raises(ValueError) as e:
        SqlAlchemyDataSampler.sample_using_random(
            execution_engine=fake_execution_engine, batch_spec=batch_spec
        )
    assert "table name must be specified" in str(e.value)


@pytest.mark.unit
def test_sample_using_random_batch_spec_test_sampling_kwargs_required():
    fake_execution_engine = None
    batch_spec = BatchSpec(table_name="table")
    with pytest.raises(ValueError) as e:
        SqlAlchemyDataSampler.sample_using_random(
            execution_engine=fake_execution_engine, batch_spec=batch_spec
        )
        assert "sample_using_random" in str(e.value)


@pytest.mark.unit
@pytest.mark.parametrize("sampling_kwargs", [{}, "a_string", []])
def test_sample_using_random_batch_spec_test_sampling_kwargs_p_required(
    sampling_kwargs,
):
    fake_execution_engine = None
    batch_spec = BatchSpec(table_name="table", sampling_kwargs=sampling_kwargs)
    with pytest.raises(ValueError) as e:
        SqlAlchemyDataSampler.sample_using_random(
            execution_engine=fake_execution_engine, batch_spec=batch_spec
        )
        assert "sample_using_random" in str(e.value)



from unittest.mock import ANY

import pandas as pd
from numpy import dtype

from great_expectations.compatibility import pyspark
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.batch.batch_column_types import (
    BatchColumnTypes,
    BatchColumnTypesResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config.databricks import (
    DatabricksDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.postgres import PostgreSQLDatasourceTestConfig
from tests.integration.test_utils.data_source_config.snowflake import SnowflakeDatasourceTestConfig
from tests.metrics.conftest import (
    PANDAS_DATA_SOURCES,
    SPARK_DATA_SOURCES,
    SQL_DATA_SOURCES,
)

DATA_FRAME = pd.DataFrame(
    {
        "numbers": [1, 2, 3],
        "strings": ["one", "two", "three"],
    },
)


@parameterize_batch_for_data_sources(
    data_source_configs=PANDAS_DATA_SOURCES,
    data=DATA_FRAME,
)
def test_pandas_success(batch_for_datasource: Batch) -> None:
    batch = batch_for_datasource
    metric = BatchColumnTypes()
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, BatchColumnTypesResult)
    assert metric_result.value == [
        {"name": "numbers", "type": dtype("int64")},
        {"name": "strings", "type": dtype("O")},
    ]


@parameterize_batch_for_data_sources(
    data_source_configs=SPARK_DATA_SOURCES,
    data=DATA_FRAME,
)
def test_spark_success(batch_for_datasource: Batch) -> None:
    batch = batch_for_datasource
    metric = BatchColumnTypes()
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, BatchColumnTypesResult)
    assert metric_result.value == [
        {"name": "numbers", "type": pyspark.types.IntegerType()},
        {"name": "strings", "type": pyspark.types.StringType()},
    ]


@parameterize_batch_for_data_sources(
    data_source_configs=[PostgreSQLDatasourceTestConfig()],
    data=DATA_FRAME,
)
def test_postgres_success(batch_for_datasource: Batch) -> None:
    batch = batch_for_datasource
    metric = BatchColumnTypes()
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, BatchColumnTypesResult)
    assert metric_result.value == [
        {"name": "numbers", "type": "INTEGER"},
        {"name": "strings", "type": "VARCHAR"},
    ]


@parameterize_batch_for_data_sources(
    data_source_configs=[DatabricksDatasourceTestConfig()],
    data=DATA_FRAME,
)
def test_databricks_success(batch_for_datasource: Batch) -> None:
    batch = batch_for_datasource
    metric = BatchColumnTypes()
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, BatchColumnTypesResult)
    assert metric_result.value == [
        {"name": "numbers", "type": "INT"},
        {"name": "strings", "type": "STRING"},
    ]


@parameterize_batch_for_data_sources(
    data_source_configs=[SnowflakeDatasourceTestConfig()],
    data=DATA_FRAME,
)
def test_snowflake_success(batch_for_datasource: Batch) -> None:
    batch = batch_for_datasource
    metric = BatchColumnTypes()
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, BatchColumnTypesResult)
    assert metric_result.value == [
        {"name": "numbers", "type": "DECIMAL(38, 0)"},
        {"name": "strings", "type": "VARCHAR(16777216)"},
    ]


@parameterize_batch_for_data_sources(
    data_source_configs=SQL_DATA_SOURCES,
    data=DATA_FRAME,
)
def test_sql_happy_path(batch_for_datasource: Batch) -> None:
    """General happy path for other SQL dialects.

    The above Postgres test validates exact types come back for sqlalchemy datasources.
    This test uses ANY to account for different SQL dialects returning different type names
    and case-insensitive string comparisons that are now used for PostgreSQL, Databricks,
    and Snowflake dialects.
    """

    batch = batch_for_datasource
    metric = BatchColumnTypes()
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, BatchColumnTypesResult)
    assert metric_result.value == [
        {"name": "numbers", "type": ANY},
        {"name": "strings", "type": ANY},
    ]



from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
)

import numpy as np

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility import sqlalchemy
from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.pyspark import pyspark
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core.metric_function_types import (
    SummarizationMetricNameSuffixes,
)
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect
from great_expectations.expectations.metrics.map_metric_provider.is_sqlalchemy_metric_selectable import (  # noqa: E501 # FIXME CoP
    _is_sqlalchemy_metric_selectable,
)
from great_expectations.expectations.metrics.util import (
    MAX_RESULT_RECORDS,
    compute_unexpected_pandas_indices,
    get_dbms_compatible_metric_domain_kwargs,
    get_sqlalchemy_source_table_and_schema,
    sql_statement_with_post_compile_to_string,
)
from great_expectations.util import (
    convert_to_json_serializable,  # noqa: TID251 # FIXME CoP
    generate_temporary_table_name,
    get_sqlalchemy_selectable,
)

if TYPE_CHECKING:
    import pandas as pd

    from great_expectations.execution_engine import (
        PandasExecutionEngine,
        SparkDFExecutionEngine,
        SqlAlchemyExecutionEngine,
    )


logger = logging.getLogger(__name__)


def _pandas_map_condition_unexpected_count(
    cls,
    execution_engine: PandasExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> int:
    """Returns unexpected count for MapExpectations"""
    return np.count_nonzero(metrics["unexpected_condition"][0])


def _pandas_map_condition_index(
    cls,
    execution_engine: PandasExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> Union[List[int], List[Dict[str, Any]]]:
    (
        boolean_mapped_unexpected_values,
        compute_domain_kwargs,
        accessor_domain_kwargs,
    ) = metrics["unexpected_condition"]

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    """
    In order to invoke the "ignore_row_if" filtering logic, "execution_engine.get_domain_records()" must be supplied
    with all of the available "domain_kwargs" keys.
    """  # noqa: E501 # FIXME CoP
    domain_kwargs = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    domain_records_df: pd.DataFrame = execution_engine.get_domain_records(
        domain_kwargs=domain_kwargs
    )
    domain_column_name_list: List[str] = list()
    # column map expectations
    if "column" in accessor_domain_kwargs:
        column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

        ###
        # NOTE: 20201111 - JPC - in the map_series / map_condition_series world (pandas), we
        # currently handle filter_column_isnull differently than other map_fn / map_condition
        # cases.
        ###
        filter_column_isnull = kwargs.get(
            "filter_column_isnull", getattr(cls, "filter_column_isnull", False)
        )
        if filter_column_isnull:
            domain_records_df = domain_records_df[domain_records_df[column_name].notnull()]

        domain_column_name_list.append(column_name)

    # column pair expectations
    elif "column_A" in accessor_domain_kwargs and "column_B" in accessor_domain_kwargs:
        column_list: List[Union[str, sqlalchemy.quoted_name]] = list()
        column_list.append(accessor_domain_kwargs["column_A"])
        column_list.append(accessor_domain_kwargs["column_B"])
        domain_column_name_list = column_list

    # multi-column map expectations
    elif "column_list" in accessor_domain_kwargs:
        column_list = accessor_domain_kwargs["column_list"]
        domain_column_name_list = column_list

    result_format = metric_value_kwargs["result_format"]
    domain_records_df = domain_records_df[boolean_mapped_unexpected_values]

    unexpected_index_list: Union[List[int], List[Dict[str, Any]]] = (
        compute_unexpected_pandas_indices(
            domain_records_df=domain_records_df,
            result_format=result_format,
            execution_engine=execution_engine,
            metrics=metrics,
            expectation_domain_column_list=domain_column_name_list,
        )
    )
    if result_format["result_format"] == "COMPLETE":
        return unexpected_index_list
    return unexpected_index_list[: result_format["partial_unexpected_count"]]


def _pandas_map_condition_query(
    cls,
    execution_engine: PandasExecutionEngine,
    metric_domain_kwargs: Dict,
    metric_value_kwargs: Dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> Optional[str]:
    """
    Returns query that will return all rows which do not meet an expected Expectation condition for instances
    of ColumnMapExpectation. For Pandas, this is currently the full set of unexpected_indices.

    Requires `unexpected_index_column_names` to be part of `result_format` dict to specify primary_key columns
    to return, along with column the Expectation is run on.
    """  # noqa: E501 # FIXME CoP
    result_format: dict = metric_value_kwargs["result_format"]

    # We will not return map_condition_query if return_unexpected_index_query = False
    return_unexpected_index_query: Optional[bool] = result_format.get(
        "return_unexpected_index_query"
    )
    if return_unexpected_index_query is False:
        return None

    (
        boolean_mapped_unexpected_values,
        compute_domain_kwargs,
        accessor_domain_kwargs,
    ) = metrics["unexpected_condition"]

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    """
    In order to invoke the "ignore_row_if" filtering logic, "execution_engine.get_domain_records()" must be supplied
    with all of the available "domain_kwargs" keys.
    """  # noqa: E501 # FIXME CoP
    domain_kwargs = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    domain_records_df: pd.DataFrame = execution_engine.get_domain_records(
        domain_kwargs=domain_kwargs
    )

    if "column" in accessor_domain_kwargs:
        column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

        filter_column_isnull = kwargs.get(
            "filter_column_isnull", getattr(cls, "filter_column_isnull", False)
        )
        if filter_column_isnull:
            domain_records_df = domain_records_df[domain_records_df[column_name].notnull()]

    domain_values_df_filtered = domain_records_df[boolean_mapped_unexpected_values]
    index_list = domain_values_df_filtered.index.to_list()
    return f"df.filter(items={index_list}, axis=0)"


def _pandas_map_condition_rows(
    cls,
    execution_engine: PandasExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> pd.DataFrame:
    """Return values from the specified domain (ignoring the column constraint) that match the map-style metric in the metrics dictionary."""  # noqa: E501 # FIXME CoP
    (
        boolean_mapped_unexpected_values,
        compute_domain_kwargs,
        accessor_domain_kwargs,
    ) = metrics["unexpected_condition"]

    accessor_domain_kwargs = get_dbms_compatible_metric_domain_kwargs(
        metric_domain_kwargs=accessor_domain_kwargs,
        batch_columns_list=metrics["table.columns"],
    )

    """
    In order to invoke the "ignore_row_if" filtering logic, "execution_engine.get_domain_records()" must be supplied
    with all of the available "domain_kwargs" keys.
    """  # noqa: E501 # FIXME CoP
    domain_kwargs = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    df = execution_engine.get_domain_records(domain_kwargs=domain_kwargs)

    if "column" in accessor_domain_kwargs:
        column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]

        ###
        # NOTE: 20201111 - JPC - in the map_series / map_condition_series world (pandas), we
        # currently handle filter_column_isnull differently than other map_fn / map_condition
        # cases.
        ###
        filter_column_isnull = kwargs.get(
            "filter_column_isnull", getattr(cls, "filter_column_isnull", False)
        )
        if filter_column_isnull:
            df = df[df[column_name].notnull()]

    result_format = metric_value_kwargs["result_format"]

    df = df[boolean_mapped_unexpected_values]

    if result_format["result_format"] == "COMPLETE":
        return df[:MAX_RESULT_RECORDS]

    limit = min(result_format["partial_unexpected_count"], MAX_RESULT_RECORDS)
    return df.iloc[:limit]


def _sqlalchemy_map_condition_unexpected_count_aggregate_fn(
    cls,
    execution_engine: SqlAlchemyExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> tuple[Any, Any, Any]:
    """Returns unexpected count for MapExpectations"""
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]

    return (
        sa.func.sum(
            sa.case(
                (unexpected_condition, 1),
                else_=0,
            )
        ),
        compute_domain_kwargs,
        accessor_domain_kwargs,
    )


def _sqlalchemy_map_condition_unexpected_count_value(
    cls,
    execution_engine: SqlAlchemyExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> float | int:
    """Returns unexpected count for MapExpectations. This is a *value* metric, which is useful for
    when the unexpected_condition is a window function.
    """
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]
    """
    In order to invoke the "ignore_row_if" filtering logic, "execution_engine.get_domain_records()" must be supplied
    with all of the available "domain_kwargs" keys.
    """  # noqa: E501 # FIXME CoP
    domain_kwargs = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    selectable = execution_engine.get_domain_records(domain_kwargs=domain_kwargs)

    # The integral values are cast to SQL Numeric in order to avoid a bug in AWS Redshift (converted to integer later).  # noqa: E501 # FIXME CoP
    count_case_statement: List[sqlalchemy.Label] = sa.case(  # type: ignore[assignment] # FIXME CoP
        (
            unexpected_condition,
            sa.sql.expression.cast(1, sa.Numeric),
        ),
        else_=sa.sql.expression.cast(0, sa.Numeric),
    ).label("condition")

    count_selectable: sqlalchemy.Select = sa.select(count_case_statement)  # type: ignore[call-overload] # FIXME CoP
    if not _is_sqlalchemy_metric_selectable(map_metric_provider=cls):
        selectable = get_sqlalchemy_selectable(selectable)  # type: ignore[arg-type] # FIXME CoP
        count_selectable = count_selectable.select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

    try:
        if execution_engine.dialect_name == GXSqlDialect.MSSQL:
            with execution_engine.get_connection() as connection:
                if not connection.closed:
                    temp_table_obj = _generate_temp_table(
                        connection=connection,
                        metric_domain_kwargs=metric_domain_kwargs,
                        metric_value_kwargs=metric_value_kwargs,
                        metrics=metrics,
                    )
                else:
                    with connection.begin():
                        temp_table_obj = _generate_temp_table(
                            connection=connection,
                            metric_domain_kwargs=metric_domain_kwargs,
                            metric_value_kwargs=metric_value_kwargs,
                            metrics=metrics,
                        )
            inner_case_query: sqlalchemy.Insert = temp_table_obj.insert().from_select(
                [count_case_statement],  # type: ignore[list-item] # FIXME CoP
                count_selectable,
            )
            execution_engine.execute_query_in_transaction(inner_case_query)  # type: ignore[arg-type] # FIXME CoP

            count_selectable = temp_table_obj  # type: ignore[assignment] # FIXME CoP

        count_selectable = get_sqlalchemy_selectable(count_selectable)  # type: ignore[assignment] # FIXME CoP
        unexpected_count_query: sqlalchemy.Select = (
            sa.select(  # type: ignore[assignment] # FIXME CoP
                sa.func.sum(sa.column("condition")).label("unexpected_count"),
            )
            .select_from(count_selectable)  # type: ignore[arg-type] # FIXME CoP
            .alias("UnexpectedCountSubquery")
        )
        unexpected_count: Union[float, int] = execution_engine.execute_query(  # type: ignore[assignment] # FIXME CoP
            sa.select(
                unexpected_count_query.c[
                    f"{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}"
                ],
            )
        ).scalar()
        # Unexpected count can be None if the table is empty, in which case the count
        # should default to zero.
        try:
            unexpected_count = int(unexpected_count)
        except TypeError:
            unexpected_count = 0

    except sqlalchemy.OperationalError as oe:
        exception_message: str = f"An SQL execution Exception occurred: {oe!s}."
        raise gx_exceptions.InvalidMetricAccessorDomainKwargsKeyError(message=exception_message)

    return convert_to_json_serializable(unexpected_count)


def _sqlalchemy_map_condition_rows(
    cls,
    execution_engine: SqlAlchemyExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> Union[Sequence[sa.Row[Any]], Any]:
    """
    Returns all rows of the metric values which do not meet an expected Expectation condition for instances
    of ColumnMapExpectation.
    """  # noqa: E501 # FIXME CoP
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]
    """
    In order to invoke the "ignore_row_if" filtering logic, "execution_engine.get_domain_records()" must be supplied
    with all of the available "domain_kwargs" keys.
    """  # noqa: E501 # FIXME CoP
    domain_kwargs = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    selectable = execution_engine.get_domain_records(domain_kwargs=domain_kwargs)

    table_columns: list[str] = metrics["table.columns"]
    column_selector = [sa.column(column_name) for column_name in table_columns]  # type: ignore[var-annotated] # FIXME CoP
    query = sa.select(*column_selector).where(unexpected_condition)
    if not _is_sqlalchemy_metric_selectable(map_metric_provider=cls):
        selectable = get_sqlalchemy_selectable(selectable)  # type: ignore[arg-type] # FIXME CoP
        query = query.select_from(selectable)  # type: ignore[arg-type] # FIXME CoP

    result_format = metric_value_kwargs["result_format"]
    if result_format["result_format"] != "COMPLETE":
        limit = min(result_format["partial_unexpected_count"], MAX_RESULT_RECORDS)
        query = query.limit(limit)
    try:
        return execution_engine.execute_query(query).fetchmany(MAX_RESULT_RECORDS)
    except sqlalchemy.OperationalError as oe:
        exception_message: str = f"An SQL execution Exception occurred: {oe!s}."
        raise gx_exceptions.InvalidMetricAccessorDomainKwargsKeyError(message=exception_message)


def _sqlalchemy_map_condition_query(  # noqa: C901 #  too complex
    cls,
    execution_engine: SqlAlchemyExecutionEngine,
    metric_domain_kwargs: Dict,
    metric_value_kwargs: Dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> Optional[str]:
    """
    Returns query that will return all rows which do not meet an expected Expectation condition for instances
    of ColumnMapExpectation.

    Requires `unexpected_index_column_names` to be part of `result_format` dict to specify primary_key columns
    to return, along with column the Expectation is run on.
    """  # noqa: E501 # FIXME CoP
    (
        unexpected_condition,
        _compute_domain_kwargs,
        accessor_domain_kwargs,
    ) = metrics["unexpected_condition"]

    result_format: dict = metric_value_kwargs["result_format"]
    # We will not return map_condition_query if return_unexpected_index_query = False
    return_unexpected_index_query: Optional[bool] = result_format.get(
        "return_unexpected_index_query"
    )
    if return_unexpected_index_query is False:
        return None

    domain_column_name_list: List[str] = list()
    # column map expectations
    if "column" in accessor_domain_kwargs:
        column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]
        domain_column_name_list.append(column_name)
    # multi-column map expectations
    elif "column_list" in accessor_domain_kwargs:
        column_list: List[Union[str, sqlalchemy.quoted_name]] = accessor_domain_kwargs[
            "column_list"
        ]
        domain_column_name_list = column_list
    # column-map expectations
    elif "column_A" in accessor_domain_kwargs and "column_B" in accessor_domain_kwargs:
        column_list = list()
        column_list.append(accessor_domain_kwargs["column_A"])
        column_list.append(accessor_domain_kwargs["column_B"])
        domain_column_name_list = column_list

    column_selector: List[sa.Column] = []

    all_table_columns: List[str] = metrics.get("table.columns", [])
    unexpected_index_column_names: List[str] | None = result_format.get(
        "unexpected_index_column_names"
    )
    if unexpected_index_column_names:
        for column_name in unexpected_index_column_names:
            if column_name not in all_table_columns:
                raise gx_exceptions.InvalidMetricAccessorDomainKwargsKeyError(
                    message=f'Error: The unexpected_index_column: "{column_name}" in does not exist in SQL Table. '  # noqa: E501 # FIXME CoP
                    f"Please check your configuration and try again."
                )

            column_selector.append(sa.column(column_name))  # type: ignore[arg-type] # FIXME CoP

    for column_name in domain_column_name_list:
        column_selector.append(sa.column(column_name))  # type: ignore[arg-type] # FIXME CoP

    unexpected_condition_query_with_selected_columns: sa.select = sa.select(*column_selector).where(  # type: ignore[valid-type] # FIXME CoP
        unexpected_condition
    )
    source_table_and_schema: sa.Table = get_sqlalchemy_source_table_and_schema(execution_engine)

    source_table_and_schema_as_selectable: Union[sa.Table, sa.Select] = get_sqlalchemy_selectable(
        source_table_and_schema
    )
    final_select_statement: sa.select = (  # type: ignore[valid-type] # FIXME CoP
        unexpected_condition_query_with_selected_columns.select_from(  # type: ignore[attr-defined] # FIXME CoP
            source_table_and_schema_as_selectable
        )
    )

    query_as_string: str = sql_statement_with_post_compile_to_string(
        engine=execution_engine, select_statement=final_select_statement
    )
    return query_as_string


def _sqlalchemy_map_condition_index(  # noqa: C901 #  too complex
    cls,
    execution_engine: SqlAlchemyExecutionEngine,
    metric_domain_kwargs: Dict,
    metric_value_kwargs: Dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> list[dict[str, Any]] | None:
    """
    Returns indices of the metric values which do not meet an expected Expectation condition for instances
    of ColumnMapExpectation.

    Requires `unexpected_index_column_names` to be part of `result_format` dict to specify primary_key columns
    to return.
    """  # noqa: E501 # FIXME CoP
    (
        unexpected_condition,
        compute_domain_kwargs,
        accessor_domain_kwargs,
    ) = metrics["unexpected_condition"]

    result_format = metric_value_kwargs["result_format"]
    if "unexpected_index_column_names" not in result_format:
        return None

    domain_column_name_list: List[str] = list()
    # column map expectations
    if "column" in accessor_domain_kwargs:
        column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]
        domain_column_name_list.append(column_name)
    # multi-column map expectations
    elif "column_list" in accessor_domain_kwargs:
        column_list: List[Union[str, sqlalchemy.quoted_name]] = accessor_domain_kwargs[
            "column_list"
        ]
        domain_column_name_list = column_list
    # column-map expectations
    elif "column_A" in accessor_domain_kwargs and "column_B" in accessor_domain_kwargs:
        column_list = list()
        column_list.append(accessor_domain_kwargs["column_A"])
        column_list.append(accessor_domain_kwargs["column_B"])
        domain_column_name_list = column_list

    domain_kwargs: dict = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    all_table_columns: List[str] = metrics.get("table.columns", [])

    unexpected_index_column_names: List[str] = result_format.get(
        "unexpected_index_column_names", []
    )

    column_selector: List[sa.Column] = []
    for column_name in unexpected_index_column_names:
        if column_name not in all_table_columns:
            raise gx_exceptions.InvalidMetricAccessorDomainKwargsKeyError(
                message=f'Error: The unexpected_index_column: "{column_name}" in does not exist in SQL Table. '  # noqa: E501 # FIXME CoP
                f"Please check your configuration and try again."
            )
        column_selector.append(sa.column(column_name))  # type: ignore[arg-type] # FIXME CoP

    # the last column we SELECT is the column the Expectation is being run on
    for column_name in domain_column_name_list:
        column_selector.append(sa.column(column_name))  # type: ignore[arg-type] # FIXME CoP

    domain_records_as_selectable: sa.sql.Selectable = execution_engine.get_domain_records(
        domain_kwargs=domain_kwargs
    )
    unexpected_condition_query_with_selected_columns: sa.select = sa.select(*column_selector).where(  # type: ignore[valid-type] # FIXME CoP
        unexpected_condition
    )

    if not _is_sqlalchemy_metric_selectable(map_metric_provider=cls):
        domain_records_as_selectable = get_sqlalchemy_selectable(domain_records_as_selectable)  # type: ignore[arg-type] # FIXME CoP

    # since SQL tables can be **very** large, truncate query_result values at 20, or at `partial_unexpected_count`  # noqa: E501 # FIXME CoP
    final_query: sa.select = unexpected_condition_query_with_selected_columns.select_from(  # type: ignore[valid-type,attr-defined] # FIXME CoP
        domain_records_as_selectable
    ).limit(result_format["partial_unexpected_count"])
    query_result: List[sqlalchemy.Row] = execution_engine.execute_query(final_query).fetchall()  # type: ignore[assignment] # FIXME CoP

    exclude_unexpected_values: bool = result_format.get("exclude_unexpected_values", False)

    return _get_sqlalchemy_customized_unexpected_index_list(
        exclude_unexpected_values=exclude_unexpected_values,
        unexpected_index_column_names=unexpected_index_column_names,
        query_result=query_result,
        domain_column_name_list=domain_column_name_list,
    )


def _spark_map_condition_unexpected_count_aggregate_fn(
    cls,
    execution_engine: SparkDFExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> tuple[Any, Any, Any]:
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]
    return (
        F.sum(F.when(unexpected_condition, 1).otherwise(0)),
        compute_domain_kwargs,
        accessor_domain_kwargs,
    )


def _spark_map_condition_unexpected_count_value(
    cls,
    execution_engine: SparkDFExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> int:
    # fn_domain_kwargs maybe updated to reflect null filtering
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]
    """
    In order to invoke the "ignore_row_if" filtering logic, "execution_engine.get_domain_records()" must be supplied
    with all of the available "domain_kwargs" keys.
    """  # noqa: E501 # FIXME CoP
    domain_kwargs = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    df = execution_engine.get_domain_records(domain_kwargs=domain_kwargs)

    # withColumn is required to transform window functions returned by some metrics to boolean mask
    data = df.withColumn("__unexpected", unexpected_condition)
    filtered = data.filter(F.col("__unexpected") == True).drop(  # noqa: E712 # FIXME CoP
        F.col("__unexpected")
    )

    return filtered.count()


def _spark_map_condition_rows(
    cls,
    execution_engine: SparkDFExecutionEngine,
    metric_domain_kwargs: dict,
    metric_value_kwargs: dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> list[pyspark.Row]:
    unexpected_condition, compute_domain_kwargs, accessor_domain_kwargs = metrics[
        "unexpected_condition"
    ]
    """
    In order to invoke the "ignore_row_if" filtering logic, "execution_engine.get_domain_records()" must be supplied
    with all of the available "domain_kwargs" keys.
    """  # noqa: E501 # FIXME CoP
    domain_kwargs = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    df = execution_engine.get_domain_records(domain_kwargs=domain_kwargs)

    # withColumn is required to transform window functions returned by some metrics to boolean mask
    data = df.withColumn("__unexpected", unexpected_condition)
    filtered = data.filter(F.col("__unexpected") == True).drop(  # noqa: E712 # FIXME CoP
        F.col("__unexpected")
    )

    result_format = metric_value_kwargs["result_format"]

    if result_format["result_format"] == "COMPLETE":
        return filtered.limit(MAX_RESULT_RECORDS).collect()

    limit = min(result_format["partial_unexpected_count"], MAX_RESULT_RECORDS)
    return filtered.limit(limit).collect()


def _spark_map_condition_index(  # noqa: C901 #  too complex
    cls,
    execution_engine: SparkDFExecutionEngine,
    metric_domain_kwargs: Dict,
    metric_value_kwargs: Dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> Union[List[Dict[str, Any]], None]:
    """
    Returns indices of the metric values which do not meet an expected Expectation condition for instances
    of ColumnMapExpectation.

    Requires `unexpected_index_column_names` to be part of `result_format` dict to specify primary_key columns
    to return.
    """  # noqa: E501 # FIXME CoP
    (
        unexpected_condition,
        compute_domain_kwargs,
        accessor_domain_kwargs,
    ) = metrics.get("unexpected_condition", (None, None, None))

    if unexpected_condition is None:
        return None

    result_format = metric_value_kwargs["result_format"]
    if "unexpected_index_column_names" not in result_format:
        return None

    domain_column_name_list: List[str] = list()
    # column map expectations
    if "column" in accessor_domain_kwargs:
        column_name: Union[str, sqlalchemy.quoted_name] = accessor_domain_kwargs["column"]
        domain_column_name_list.append(column_name)

    # multi-column map expectations
    elif "column_list" in accessor_domain_kwargs:
        column_list: List[Union[str, sqlalchemy.quoted_name]] = accessor_domain_kwargs[
            "column_list"
        ]
        domain_column_name_list = column_list
    # column-map expectations
    elif "column_A" in accessor_domain_kwargs and "column_B" in accessor_domain_kwargs:
        column_list = list()
        column_list.append(accessor_domain_kwargs["column_A"])
        column_list.append(accessor_domain_kwargs["column_B"])
        domain_column_name_list = column_list

    domain_kwargs = dict(**compute_domain_kwargs, **accessor_domain_kwargs)
    df: pyspark.sql.dataframe.DataFrame = execution_engine.get_domain_records(
        domain_kwargs=domain_kwargs
    )
    result_format = metric_value_kwargs["result_format"]
    if not result_format.get("unexpected_index_column_names"):
        raise gx_exceptions.MetricResolutionError(
            message="unexpected_indices cannot be returned without 'unexpected_index_column_names'. Please check your configuration.",  # noqa: E501 # FIXME CoP
            failed_metrics=["unexpected_index_list"],
        )
    # withColumn is required to transform window functions returned by some metrics to boolean mask
    data = df.withColumn("__unexpected", unexpected_condition)
    filtered = data.filter(F.col("__unexpected") == True).drop(  # noqa: E712 # FIXME CoP
        F.col("__unexpected")
    )
    exclude_unexpected_values: bool = result_format.get("exclude_unexpected_values", False)

    unexpected_index_column_names: List[str] = result_format["unexpected_index_column_names"]
    columns_to_keep: List[str] = [column for column in unexpected_index_column_names]
    columns_to_keep += domain_column_name_list

    # check that column name is in row
    for col_name in columns_to_keep:
        if col_name not in filtered.columns:
            raise gx_exceptions.InvalidMetricAccessorDomainKwargsKeyError(  # noqa: TRY003 # FIXME CoP
                f"Error: The unexpected_index_column '{col_name}' does not exist in Spark DataFrame. Please check your configuration and try again."  # noqa: E501 # FIXME CoP
            )

    if result_format["result_format"] != "COMPLETE":
        filtered = filtered.limit(result_format["partial_unexpected_count"])

    # Prune the dataframe down only the columns we care about
    filtered = filtered.select(columns_to_keep)

    return _get_spark_customized_unexpected_index_list(
        exclude_unexpected_values=exclude_unexpected_values,
        unexpected_index_column_names=unexpected_index_column_names,
        filtered=filtered,
        columns_to_keep=columns_to_keep,
    )


def _spark_map_condition_query(
    cls,
    execution_engine: SparkDFExecutionEngine,
    metric_domain_kwargs: Dict,
    metric_value_kwargs: Dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> Union[str, None]:
    """
    Returns query that will return all rows which do not meet an expected Expectation condition for instances
    of ColumnMapExpectation.

    Converts unexpected_condition into a string that can be rendered in DataDocs

    Output will look like:

        df.filter(F.expr( [unexpected_condition] ))

    """  # noqa: E501 # FIXME CoP
    result_format: dict = metric_value_kwargs["result_format"]
    # We will not return map_condition_query if return_unexpected_index_query = False
    return_unexpected_index_query: Optional[bool] = result_format.get(
        "return_unexpected_index_query"
    )
    if return_unexpected_index_query is False:
        return None

    (
        unexpected_condition,
        _,
        _,
    ) = metrics.get("unexpected_condition", (None, None, None))

    # unexpected_condition is an F.column object, meaning the str representation is wrapped in Column<> syntax.  # noqa: E501 # FIXME CoP
    # like Column<'[unexpected_expression]'>
    unexpected_condition_as_string: str = str(unexpected_condition)
    unexpected_condition_filtered: str = unexpected_condition_as_string.replace(
        "Column<'(", ""
    ).replace(")'>", "")
    return f"df.filter(F.expr({unexpected_condition_filtered}))"


def _generate_temp_table(
    connection: sa.engine.base.Connection,
    metric_domain_kwargs: Dict,
    metric_value_kwargs: Dict,
    metrics: Dict[str, Any],
    **kwargs,
) -> sa.Table:
    temp_table_name: str = generate_temporary_table_name(default_table_name_prefix="#ge_temp_")
    metadata: sa.MetaData = sa.MetaData()
    metadata.reflect(bind=connection)
    temp_table_obj: sa.Table = sa.Table(
        temp_table_name,
        metadata,
        sa.Column("condition", sa.Integer, primary_key=False, nullable=False),
    )
    temp_table_obj.create(bind=connection, checkfirst=True)
    return temp_table_obj


def _get_sqlalchemy_customized_unexpected_index_list(
    exclude_unexpected_values: bool,
    unexpected_index_column_names: List[str],
    query_result: List[sqlalchemy.Row],
    domain_column_name_list: List[Union[str, sqlalchemy.quoted_name]],
) -> Union[List[Dict[str, Any]], None]:
    unexpected_index_list: List[Dict[str, Any]] = []

    if (
        exclude_unexpected_values
        and len(query_result) != 0
        and len(unexpected_index_column_names) != 0
    ):
        primary_key_dict_list: dict[str, List[Any]] = {
            idx_col: [] for idx_col in unexpected_index_column_names
        }
        for row in query_result:
            for index in range(len(unexpected_index_column_names)):
                primary_key_dict_list[unexpected_index_column_names[index]].append(row[index])
        unexpected_index_list.append(primary_key_dict_list)

    else:
        for row in query_result:
            primary_key_dict: Dict[str, Any] = {}
            # add the actual unexpected value
            all_columns = unexpected_index_column_names + domain_column_name_list
            for index in range(len(all_columns)):
                name: str = all_columns[index]
                primary_key_dict[name] = row[index]
            unexpected_index_list.append(primary_key_dict)

    return unexpected_index_list


def _get_spark_customized_unexpected_index_list(
    exclude_unexpected_values: bool,
    unexpected_index_column_names: List[str],
    filtered: pyspark.sql.dataframe.DataFrame,
    columns_to_keep: List[str],
) -> Union[List[Dict[str, Any]], None]:
    unexpected_index_list: List[Dict[str, Any]] = []

    if exclude_unexpected_values and not filtered.isEmpty():
        dict_list_to_add: dict[str, List[Any]] = {
            idx_col: [] for idx_col in unexpected_index_column_names
        }
        for row in filtered.collect():
            for col_name in unexpected_index_column_names:
                dict_list_to_add[col_name].append(row[col_name])
        unexpected_index_list.append(dict_list_to_add)

    else:
        for row in filtered.collect():
            dict_to_add: dict = {}
            for col_name in columns_to_keep:
                dict_to_add[col_name] = row[col_name]
            unexpected_index_list.append(dict_to_add)

    return unexpected_index_list



from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import Extra


class Offset(pydantic.BaseModel):
    """
    A threshold in which a metric will be considered passable
    """

    positive: float
    negative: float

    class Config:
        extra = Extra.forbid


class Window(pydantic.BaseModel):
    """
    A definition for a temporal window across <`range`> number of previous invocations
    """

    constraint_fn: str
    parameter_name: str
    range: int
    offset: Offset
    strict: bool = False

    class Config:
        extra = Extra.forbid



from __future__ import annotations

import inspect
import re
from collections import defaultdict
from dataclasses import asdict, dataclass
from typing import List, Sequence, Tuple, Union

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.expectation_diagnostics.expectation_test_data_cases import (
    ExpectationTestDataCases,  # noqa: TC001 # FIXME CoP
)
from great_expectations.core.expectation_diagnostics.supporting_types import (
    AugmentedLibraryMetadata,
    ExpectationBackendTestResultCounts,
    ExpectationDescriptionDiagnostics,
    ExpectationDiagnosticCheckMessage,
    ExpectationDiagnosticCheckMessageDict,
    ExpectationDiagnosticMaturityMessages,
    ExpectationErrorDiagnostics,
    ExpectationExecutionEngineDiagnostics,
    ExpectationMetricDiagnostics,
    ExpectationRendererDiagnostics,
    ExpectationTestDiagnostics,
)
from great_expectations.exceptions import InvalidExpectationConfigurationError
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.types import SerializableDictDot
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP


@dataclass(frozen=True)
class ExpectationDiagnostics(SerializableDictDot):
    """An immutable object created by Expectation.run_diagnostics.
    It contains information introspected from the Expectation class, in formats that can be renderered at the command line, and by the Gallery.

    It has three external-facing use cases:

    1. `ExpectationDiagnostics.to_dict()` creates the JSON object that populates the Gallery.
    2. `ExpectationDiagnostics.generate_checklist()` creates CLI-type string output to assist with development.
    """  # noqa: E501 # FIXME CoP

    # This object is taken directly from the Expectation class, without modification
    examples: List[ExpectationTestDataCases]
    gallery_examples: List[ExpectationTestDataCases]

    # These objects are derived from the Expectation class
    # They're a combination of direct introspection of existing properties,
    # and instantiating the Expectation with test data and actually executing
    # methods.
    # For example, we can verify the existence of certain Renderers through
    # introspection alone, but in order to see what they return, we need to
    # instantiate the Expectation and actually run the method.

    library_metadata: Union[AugmentedLibraryMetadata, ExpectationDescriptionDiagnostics]
    description: ExpectationDescriptionDiagnostics
    execution_engines: ExpectationExecutionEngineDiagnostics

    renderers: List[ExpectationRendererDiagnostics]
    metrics: List[ExpectationMetricDiagnostics]
    tests: List[ExpectationTestDiagnostics]
    backend_test_result_counts: List[ExpectationBackendTestResultCounts]
    errors: List[ExpectationErrorDiagnostics]
    maturity_checklist: ExpectationDiagnosticMaturityMessages
    coverage_score: float

    @override
    def to_json_dict(self) -> dict:
        result = convert_to_json_serializable(data=asdict(self))
        result["execution_engines_list"] = sorted(
            [engine for engine, _bool in result["execution_engines"].items() if _bool is True]
        )
        return result

    def generate_checklist(self) -> str:
        """Generates the checklist in CLI-appropriate string format."""
        str_ = self._convert_checks_into_output_message(
            self.description["camel_name"],
            self.library_metadata.maturity,  # type: ignore[union-attr] # could be ExpectationDescriptionDiagnostics
            self.maturity_checklist,
        )
        return str_

    @staticmethod
    def _check_library_metadata(
        library_metadata: Union[AugmentedLibraryMetadata, ExpectationDescriptionDiagnostics],
    ) -> ExpectationDiagnosticCheckMessage:
        """Check whether the Expectation has a library_metadata object"""
        sub_messages: list[ExpectationDiagnosticCheckMessageDict] = []
        for problem in library_metadata.problems:  # type: ignore[union-attr] # could be ExpectationDescriptionDiagnostics
            sub_messages.append(
                {
                    "message": problem,
                    "passed": False,
                }
            )

        return ExpectationDiagnosticCheckMessage(
            message="Has a valid library_metadata object",
            passed=library_metadata.library_metadata_passed_checks,  # type: ignore[union-attr] # could be ExpectationDescriptionDiagnostics
            sub_messages=sub_messages,
        )

    @staticmethod
    def _check_docstring(
        description: ExpectationDescriptionDiagnostics,
    ) -> ExpectationDiagnosticCheckMessage:
        """Check whether the Expectation has an informative docstring"""

        message = 'Has a docstring, including a one-line short description that begins with "Expect" and ends with a period'  # noqa: E501 # FIXME CoP
        if "short_description" in description:
            short_description = description["short_description"]
        else:
            short_description = None
        if short_description in {"", "\n", "TODO: Add a docstring here", None}:
            return ExpectationDiagnosticCheckMessage(
                message=message,
                passed=False,
            )
        elif short_description.startswith("Expect ") and short_description.endswith("."):
            return ExpectationDiagnosticCheckMessage(
                message=message,
                sub_messages=[
                    {
                        "message": f'"{short_description}"',
                        "passed": True,
                    }
                ],
                passed=True,
            )
        else:
            return ExpectationDiagnosticCheckMessage(
                message=message,
                sub_messages=[
                    {
                        "message": f'"{short_description}"',
                        "passed": False,
                    }
                ],
                passed=False,
            )

    @classmethod
    def _check_example_cases(
        cls,
        examples: List[ExpectationTestDataCases],
        tests: List[ExpectationTestDiagnostics],
    ) -> ExpectationDiagnosticCheckMessage:
        """Check whether this Expectation has at least one positive and negative example case (and all test cases return the expected output)"""  # noqa: E501 # FIXME CoP

        message = "Has at least one positive and negative example case, and all test cases pass"
        (
            positive_case_count,
            negative_case_count,
        ) = cls._count_positive_and_negative_example_cases(examples)
        unexpected_case_count = cls._count_unexpected_test_cases(tests)
        passed = (
            (positive_case_count > 0) and (negative_case_count > 0) and (unexpected_case_count == 0)
        )
        print(positive_case_count, negative_case_count, unexpected_case_count, passed)
        return ExpectationDiagnosticCheckMessage(
            message=message,
            passed=passed,
        )

    @staticmethod
    def _check_core_logic_for_at_least_one_execution_engine(
        backend_test_result_counts: List[ExpectationBackendTestResultCounts],
    ) -> ExpectationDiagnosticCheckMessage:
        """Check whether core logic for this Expectation exists and passes tests on at least one Execution Engine"""  # noqa: E501 # FIXME CoP

        sub_messages: List[ExpectationDiagnosticCheckMessageDict] = []
        passed = False
        message = "Has core logic and passes tests on at least one Execution Engine"
        all_passing = [
            backend_test_result
            for backend_test_result in backend_test_result_counts
            if backend_test_result.failing_names is None and backend_test_result.num_passed >= 1
        ]

        if len(all_passing) > 0:
            passed = True
            for result in all_passing:
                sub_messages.append(
                    {
                        "message": f"All {result.num_passed} tests for {result.backend} are passing",  # noqa: E501 # FIXME CoP
                        "passed": True,
                    }
                )

        if not backend_test_result_counts:
            sub_messages.append(
                {
                    "message": "There are no test results",
                    "passed": False,
                }
            )

        return ExpectationDiagnosticCheckMessage(
            message=message,
            passed=passed,
            sub_messages=sub_messages,
        )

    @staticmethod
    def _get_backends_from_test_results(
        test_results: List[ExpectationTestDiagnostics],
    ) -> List[ExpectationBackendTestResultCounts]:
        """Has each tested backend and the number of passing/failing tests"""
        backend_results = defaultdict(list)
        backend_failing_names = defaultdict(list)
        results: List[ExpectationBackendTestResultCounts] = []

        for test_result in test_results:
            backend_results[test_result.backend].append(test_result.test_passed)
            if test_result.test_passed is False:
                backend_failing_names[test_result.backend].append(test_result.test_title)

        for backend in backend_results:
            result_counts = ExpectationBackendTestResultCounts(
                backend=backend,
                num_passed=backend_results[backend].count(True),
                num_failed=backend_results[backend].count(False),
                failing_names=backend_failing_names.get(backend),
            )
            results.append(result_counts)

        return results

    @staticmethod
    def _check_core_logic_for_all_applicable_execution_engines(
        backend_test_result_counts: List[ExpectationBackendTestResultCounts],
    ) -> ExpectationDiagnosticCheckMessage:
        """Check whether core logic for this Expectation exists and passes tests on all applicable Execution Engines"""  # noqa: E501 # FIXME CoP

        sub_messages: list[ExpectationDiagnosticCheckMessageDict] = []
        passed = False
        message = (
            "Has core logic that passes tests for all applicable Execution Engines and SQL dialects"
        )
        all_passing = [
            backend_test_result
            for backend_test_result in backend_test_result_counts
            if backend_test_result.failing_names is None and backend_test_result.num_passed >= 1
        ]
        some_failing = [
            backend_test_result
            for backend_test_result in backend_test_result_counts
            if backend_test_result.failing_names is not None
        ]

        if len(all_passing) > 0 and len(some_failing) == 0:
            passed = True

        for result in all_passing:
            sub_messages.append(
                {
                    "message": f"All {result.num_passed} tests for {result.backend} are passing",
                    "passed": True,
                }
            )

        for result in some_failing:
            sub_messages.append(
                {
                    "message": f"Only {result.num_passed} / {result.num_passed + result.num_failed} tests for {result.backend} are passing",  # noqa: E501 # FIXME CoP
                    "passed": False,
                }
            )
            sub_messages.append(
                {
                    "message": f"  - Failing: {', '.join(result.failing_names)}",  # type: ignore[arg-type] # FIXME CoP
                    "passed": False,
                }
            )

        if not backend_test_result_counts:
            sub_messages.append(
                {
                    "message": "There are no test results",
                    "passed": False,
                }
            )

        return ExpectationDiagnosticCheckMessage(
            message=message,
            passed=passed,
            sub_messages=sub_messages,
        )

    @staticmethod
    def _count_positive_and_negative_example_cases(
        examples: List[ExpectationTestDataCases],
    ) -> Tuple[int, int]:
        """Scans examples and returns a 2-ple with the numbers of cases with success == True and success == False"""  # noqa: E501 # FIXME CoP

        positive_cases: int = 0
        negative_cases: int = 0

        for test_data_cases in examples:
            for test in test_data_cases["tests"]:
                success = test["output"].get("success")
                if success is True:
                    positive_cases += 1
                elif success is False:
                    negative_cases += 1
        return positive_cases, negative_cases

    @staticmethod
    def _count_unexpected_test_cases(
        test_diagnostics: Sequence[ExpectationTestDiagnostics],
    ) -> int:
        """Scans test_diagnostics and returns the number of cases that did not pass."""

        unexpected_cases: int = 0

        for test in test_diagnostics:
            passed = test["test_passed"] is True
            if not passed:
                unexpected_cases += 1

        return unexpected_cases

    @staticmethod
    def _convert_checks_into_output_message(
        class_name: str,
        maturity_level: str,
        maturity_messages: ExpectationDiagnosticMaturityMessages,
    ) -> str:
        """Converts a list of checks into an output string (potentially nested), with ✔ to indicate checks that passed."""  # noqa: E501 # FIXME CoP

        output_message = f"Completeness checklist for {class_name} ({maturity_level}):"

        checks = (
            maturity_messages.experimental + maturity_messages.beta + maturity_messages.production
        )

        for check in checks:
            if check["passed"]:
                output_message += f"\n ✔ {check['message']}"
            else:
                output_message += f"\n   {check['message']}"

            if "sub_messages" in check:
                for sub_message in check["sub_messages"]:
                    if sub_message["passed"]:
                        output_message += f"\n    ✔ {sub_message['message']}"
                    else:
                        output_message += f"\n      {sub_message['message']}"
        output_message += "\n"

        return output_message

    @staticmethod
    def _check_input_validation(
        expectation_instance,
        examples: List[ExpectationTestDataCases],
    ) -> ExpectationDiagnosticCheckMessage:
        """Check that the validate_configuration exists and doesn't raise a config error"""
        passed = False
        sub_messages: list[ExpectationDiagnosticCheckMessageDict] = []
        rx = re.compile(r"^[\s]+assert", re.MULTILINE)
        try:
            first_test = examples[0]["tests"][0]
        except IndexError:
            sub_messages.append(
                {
                    "message": "No example found to get kwargs for ExpectationConfiguration",
                    "passed": passed,
                }
            )
        else:
            if "validate_configuration" not in expectation_instance.__class__.__dict__:
                sub_messages.append(
                    {
                        "message": "No validate_configuration method defined on subclass",
                        "passed": passed,
                    }
                )
            else:
                expectation_config = ExpectationConfiguration(
                    type=expectation_instance.expectation_type,
                    kwargs=first_test.input,
                )
                validate_configuration_source = inspect.getsource(
                    expectation_instance.__class__.validate_configuration
                )
                if rx.search(validate_configuration_source):
                    sub_messages.append(
                        {
                            "message": "Custom 'assert' statements in validate_configuration",
                            "passed": True,
                        }
                    )
                else:
                    sub_messages.append(
                        {
                            "message": "Using default validate_configuration from template",
                            "passed": False,
                        }
                    )
                try:
                    expectation_instance.validate_configuration(expectation_config)
                except InvalidExpectationConfigurationError:
                    pass
                else:
                    passed = True

        return ExpectationDiagnosticCheckMessage(
            message="Has basic input validation and type checking",
            passed=passed,
            sub_messages=sub_messages,
        )

    @staticmethod
    def _check_renderer_methods(
        expectation_instance,
    ) -> ExpectationDiagnosticCheckMessage:
        """Check if all statment renderers are defined"""
        passed = False
        # For now, don't include the "question", "descriptive", or "answer"
        # types since they are so sparsely implemented
        # all_renderer_types = {"diagnostic", "prescriptive", "question", "descriptive", "answer"}
        all_renderer_types = {"diagnostic", "prescriptive"}
        renderer_names = [
            name
            for name in dir(expectation_instance)
            if name.endswith("renderer") and name.startswith("_")
        ]
        renderer_types = {name.split("_")[1] for name in renderer_names}
        if all_renderer_types & renderer_types == all_renderer_types:
            passed = True
        return ExpectationDiagnosticCheckMessage(
            # message="Has all four statement Renderers: question, descriptive, prescriptive, diagnostic",  # noqa: E501 # FIXME CoP
            message="Has both statement Renderers: prescriptive and diagnostic",
            passed=passed,
        )

    @staticmethod
    def _check_full_test_suite(
        library_metadata: Union[AugmentedLibraryMetadata, ExpectationDescriptionDiagnostics],
    ) -> ExpectationDiagnosticCheckMessage:
        """Check library_metadata to see if Expectation has a full test suite"""
        return ExpectationDiagnosticCheckMessage(
            message="Has a full suite of tests, as determined by a code owner",
            passed=library_metadata.has_full_test_suite,  # type: ignore[union-attr] # could be ExpectationDescriptionDiagnostics
        )

    @staticmethod
    def _check_manual_code_review(
        library_metadata: Union[AugmentedLibraryMetadata, ExpectationDescriptionDiagnostics],
    ) -> ExpectationDiagnosticCheckMessage:
        """Check library_metadata to see if a manual code review has been performed"""
        return ExpectationDiagnosticCheckMessage(
            message="Has passed a manual review by a code owner for code standards and style guides",  # noqa: E501 # FIXME CoP
            passed=library_metadata.manually_reviewed_code,  # type: ignore[union-attr] # could be ExpectationDescriptionDiagnostics
        )



"""Note: AWS Glue split from spark since it requires different test dependencies."""

from tests.integration.backend_dependencies import BackendDependencies
from tests.integration.integration_test_fixture import IntegrationTestFixture

aws_glue_integration_tests = []

deployment_patterns = [
    # TODO: The AWS_GLUE dependency is only being marked and not run at this time.
    IntegrationTestFixture(
        name="how_to_use_great_expectations_in_aws_glue",
        user_flow_script="tests/integration/docusaurus/deployment_patterns/aws_glue_deployment_patterns.py",
        backend_dependencies=[
            BackendDependencies.SPARK,
            BackendDependencies.AWS,
            BackendDependencies.AWS_GLUE,
        ],
    ),
]

aws_glue_integration_tests += deployment_patterns



from __future__ import annotations

import logging
import pathlib
import re
from typing import TYPE_CHECKING

import boto3
import botocore
import pytest

from great_expectations.core.partitioners import FileNamePartitionerPath
from great_expectations.datasource.fluent import SparkDBFSDatasource
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import CSVAsset
from tests.test_utils import create_files_in_directory

if TYPE_CHECKING:
    from pyfakefs.fake_filesystem import FakeFilesystem


logger = logging.getLogger(__file__)


@pytest.fixture
def spark_dbfs_datasource(fs: FakeFilesystem, test_backends) -> SparkDBFSDatasource:
    if "SparkDFDataset" not in test_backends:
        pytest.skip("No spark backend selected.")

    # Copy boto modules into fake filesystem (see https://github.com/spulec/moto/issues/1682#issuecomment-645016188)
    for module in [boto3, botocore]:
        module_dir = pathlib.Path(module.__file__).parent
        fs.add_real_directory(module_dir, lazy_read=False)

    base_directory: str = "/dbfs/great_expectations"
    fs.create_dir(base_directory)

    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "alex_20200809_1000.csv",
            "eugene_20200809_1500.csv",
            "james_20200811_1009.csv",
            "abe_20200809_1040.csv",
            "will_20200809_1002.csv",
            "james_20200713_1567.csv",
            "eugene_20201129_1900.csv",
            "will_20200810_1001.csv",
            "james_20200810_1003.csv",
            "alex_20200819_1300.csv",
        ],
    )

    return SparkDBFSDatasource(
        name="spark_dbfs_datasource",
        base_directory=pathlib.Path(base_directory),
    )


@pytest.mark.spark
def test_construct_spark_dbfs_datasource(spark_dbfs_datasource: SparkDBFSDatasource):
    assert spark_dbfs_datasource.name == "spark_dbfs_datasource"


@pytest.mark.spark
def test_add_csv_asset_to_datasource(spark_dbfs_datasource: SparkDBFSDatasource):
    asset_specified_metadata = {"asset_level_metadata": "my_metadata"}
    asset = spark_dbfs_datasource.add_csv_asset(
        name="csv_asset",
        batch_metadata=asset_specified_metadata,
    )
    assert asset.name == "csv_asset"
    assert asset.batch_metadata == asset_specified_metadata


@pytest.mark.unit
def test_construct_csv_asset_directly():
    asset = CSVAsset(
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.spark
@pytest.mark.xfail(
    reason="Accessing objects on pyfakefs.fake_filesystem.FakeFilesystem using Spark is not working (this test is conducted using Jupyter notebook manually)."  # noqa: E501 # FIXME CoP
)
def test_get_batch_list_from_fully_specified_batch_request(
    spark_dbfs_datasource: SparkDBFSDatasource,
):
    asset_specified_metadata = {"asset_level_metadata": "my_metadata"}
    asset = spark_dbfs_datasource.add_csv_asset(
        name="csv_asset",
        batch_metadata=asset_specified_metadata,
    )

    batching_regex = re.compile(r"(?P<name>.+)_(?P<timestamp>.+)_(?P<price>\d{4})\.csv")
    request = asset.build_batch_request(
        options={"name": "alex", "timestamp": "20200819", "price": "1300"},
        partitioner=FileNamePartitionerPath(regex=batching_regex),
    )
    batch = asset.get_batch(request)
    assert batch.batch_request.datasource_name == spark_dbfs_datasource.name
    assert batch.batch_request.data_asset_name == asset.name
    assert batch.batch_request.options == (
        "name",
        "timestamp",
        "price",
        "path",
    )
    assert batch.metadata == {
        "path": "alex_20200819_1300.csv",
        "name": "alex",
        "timestamp": "20200819",
        "price": "1300",
        **asset_specified_metadata,
    }
    assert batch.id == "spark_dbfs_datasource-csv_asset-name_alex-timestamp_20200819-price_1300"

    request = asset.build_batch_request(
        options={"name": "alex"}, partitioner=FileNamePartitionerPath(regex=batching_regex)
    )
    assert len(asset.get_batch_identifiers_list(request)) == 2



from __future__ import annotations

from great_expectations.compatibility import pyspark
from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core.metric_function_types import MetricPartialFunctionTypes
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric_provider import (
    ColumnMapMetricProvider,
    column_condition_partial,
)
from great_expectations.util import generate_temporary_table_name


class ColumnValuesUnique(ColumnMapMetricProvider):
    condition_metric_name = "column_values.unique"

    @column_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column, **kwargs):
        return ~column.duplicated(keep=False)

    # NOTE: 20201119 - JPC - We cannot split per-dialect into window and non-window functions
    # @column_condition_partial(
    #     engine=SqlAlchemyExecutionEngine,
    # )
    # def _sqlalchemy(cls, column, _table, **kwargs):
    #     dup_query = (
    #         sa.select(column)
    #         .select_from(_table)
    #         .group_by(column)
    #         .having(sa.func.count(column) > 1)
    #     )
    #
    #     return column.notin_(dup_query)

    @column_condition_partial(
        engine=SqlAlchemyExecutionEngine,
        partial_fn_type=MetricPartialFunctionTypes.WINDOW_CONDITION_FN,
    )
    def _sqlalchemy_window(cls, column, _table, **kwargs):
        # Will - 20210126
        # This is a special case that needs to be handled for mysql, where you cannot refer to a temp_table  # noqa: E501 # FIXME CoP
        # more than once in the same query. So instead of passing dup_query as-is, a second temp_table is created with  # noqa: E501 # FIXME CoP
        # the column we will be performing the expectation on, and the query is performed against it.  # noqa: E501 # FIXME CoP
        dialect = kwargs.get("_dialect")
        sql_engine = kwargs.get("_sqlalchemy_engine")
        execution_engine = kwargs.get("_execution_engine")
        try:
            dialect_name = dialect.dialect.name
        except AttributeError:
            try:
                dialect_name = dialect.name
            except AttributeError:
                dialect_name = ""
        if sql_engine and dialect and dialect_name == "mysql":
            temp_table_name = generate_temporary_table_name()
            temp_table_stmt = f"CREATE TEMPORARY TABLE {temp_table_name} AS SELECT tmp.{column.name} FROM {_table} tmp"  # noqa: E501 # FIXME CoP
            execution_engine.execute_query_in_transaction(sa.text(temp_table_stmt))
            dup_query = (
                sa.select(column)
                .select_from(sa.text(temp_table_name))
                .group_by(column)
                .having(sa.func.count(column) > 1)
            )
        else:
            dup_query = (
                sa.select(column)
                .select_from(_table)
                .group_by(column)
                .having(sa.func.count(column) > 1)
            )
        return column.notin_(dup_query)

    @column_condition_partial(
        engine=SparkDFExecutionEngine,
        partial_fn_type=MetricPartialFunctionTypes.WINDOW_CONDITION_FN,
    )
    def _spark(cls, column, **kwargs):
        return F.count(F.lit(1)).over(pyspark.Window.partitionBy(column)) <= 1



import os
import shutil
from typing import Set, Tuple

import pytest

import great_expectations as gx
from great_expectations.core import ExpectationSuite
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.util import file_relative_path


@pytest.fixture
def titanic_data_context_modular_api(tmp_path_factory, monkeypatch):
    project_path = str(tmp_path_factory.mktemp("titanic_data_context"))
    context_path = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(context_path, "expectations"),  # noqa: PTH118 # FIXME CoP
        exist_ok=True,
    )
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(context_path, "checkpoints"),  # noqa: PTH118 # FIXME CoP
        exist_ok=True,
    )
    data_path = os.path.join(context_path, "../data")  # noqa: PTH118 # FIXME CoP
    os.makedirs(os.path.join(data_path), exist_ok=True)  # noqa: PTH118, PTH103 # FIXME CoP
    titanic_yml_path = file_relative_path(
        __file__, "./fixtures/great_expectations_titanic_0.13.yml"
    )
    shutil.copy(
        titanic_yml_path,
        str(os.path.join(context_path, FileDataContext.GX_YML)),  # noqa: PTH118 # FIXME CoP
    )
    titanic_csv_path = file_relative_path(__file__, "../test_sets/Titanic.csv")
    shutil.copy(
        titanic_csv_path,
        str(os.path.join(context_path, "../data/Titanic.csv")),  # noqa: PTH118 # FIXME CoP
    )
    return gx.get_context(context_root_dir=context_path)


@pytest.fixture()
def possible_expectations_set():
    return {
        "expect_table_columns_to_match_ordered_list",
        "expect_table_row_count_to_be_between",
        "expect_column_values_to_be_in_type_list",
        "expect_column_values_to_not_be_null",
        "expect_column_values_to_be_null",
        "expect_column_proportion_of_unique_values_to_be_between",
        "expect_column_min_to_be_between",
        "expect_column_max_to_be_between",
        "expect_column_mean_to_be_between",
        "expect_column_median_to_be_between",
        "expect_column_quantile_values_to_be_between",
        "expect_column_values_to_be_in_set",
        "expect_column_values_to_be_between",
        "expect_column_values_to_be_unique",
        "expect_compound_columns_to_be_unique",
    }


def get_set_of_columns_and_expectations_from_suite(
    suite: ExpectationSuite,
) -> Tuple[Set[str], Set[str]]:
    """
    Args:
        suite: An expectation suite

    Returns:
        A tuple containing a set of columns and a set of expectations found in a suite
    """
    columns: Set[str] = {
        i.kwargs.get("column")  # type: ignore[misc] # FIXME CoP
        for i in suite.expectation_configurations
        if i.kwargs.get("column")
    }
    expectations: Set[str] = {i.type for i in suite.expectation_configurations}

    return columns, expectations



import pytest

from great_expectations import expectations as gxe
from great_expectations.render import (
    RenderedAtomicContent,
    RenderedAtomicValue,
)
from great_expectations.render.renderer.inline_renderer import InlineRenderer


@pytest.mark.parametrize(
    "expectation,expected_expectation_configuration_rendered_atomic_content",
    [
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template="$column may have any proportion of non-null values.",
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            }
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="no_min_max_values",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                min_value=0.5,
                max_value=0.8,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template=(
                            "$column proportion of non-null values must be greater "
                            "than or equal to $min_value and less than or equal to $max_value."
                        ),
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "min_value": {
                                "schema": {"type": "number"},
                                "value": 0.5,
                            },
                            "max_value": {
                                "schema": {"type": "number"},
                                "value": 0.8,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="min_and_max_values",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                max_value=0.8,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template=(
                            "$column proportion of non-null values must be less than or equal to "
                            "$max_value."
                        ),
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "max_value": {
                                "schema": {"type": "number"},
                                "value": 0.8,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="only_max_value",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                min_value=0.5,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template=(
                            "$column proportion of non-null values must be greater "
                            "than or equal to $min_value."
                        ),
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "min_value": {
                                "schema": {"type": "number"},
                                "value": 0.5,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="only_min_value",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                min_value=0.7,
                max_value=0.7,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template="$column proportion of non-null values "
                        "must be exactly $min_value.",
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "min_value": {
                                "schema": {"type": "number"},
                                "value": 0.7,
                            },
                            "max_value": {
                                "schema": {"type": "number"},
                                "value": 0.7,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="equal_min_max_values",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                min_value=0.5,
                max_value=0.8,
                strict_min=True,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template=(
                            "$column proportion of non-null values must be greater than $min_value "
                            "and less than or equal to $max_value."
                        ),
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "min_value": {
                                "schema": {"type": "number"},
                                "value": 0.5,
                            },
                            "max_value": {
                                "schema": {"type": "number"},
                                "value": 0.8,
                            },
                            "strict_min": {
                                "schema": {"type": "boolean"},
                                "value": True,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="strict_min_condition",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                min_value=0.5,
                max_value=0.8,
                strict_max=True,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template=(
                            "$column proportion of non-null values must be greater "
                            "than or equal to $min_value and less than $max_value."
                        ),
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "min_value": {
                                "schema": {"type": "number"},
                                "value": 0.5,
                            },
                            "max_value": {
                                "schema": {"type": "number"},
                                "value": 0.8,
                            },
                            "strict_max": {
                                "schema": {"type": "boolean"},
                                "value": True,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="strict_max_condition",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                min_value=0.5,
                max_value=0.8,
                strict_min=True,
                strict_max=True,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template=(
                            "$column proportion of non-null values must be greater than $min_value "
                            "and less than $max_value."
                        ),
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "min_value": {
                                "schema": {"type": "number"},
                                "value": 0.5,
                            },
                            "max_value": {
                                "schema": {"type": "number"},
                                "value": 0.8,
                            },
                            "strict_min": {
                                "schema": {"type": "boolean"},
                                "value": True,
                            },
                            "strict_max": {
                                "schema": {"type": "boolean"},
                                "value": True,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="both_strict_conditions",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                min_value=0.5,
                strict_min=True,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template=(
                            "$column proportion of non-null values must be greater than $min_value."
                        ),
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "min_value": {
                                "schema": {"type": "number"},
                                "value": 0.5,
                            },
                            "strict_min": {
                                "schema": {"type": "boolean"},
                                "value": True,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="only_min_value_strict",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column="test_column",
                max_value=0.8,
                strict_max=True,
            ),
            [
                RenderedAtomicContent(
                    name="atomic.prescriptive.summary",
                    value=RenderedAtomicValue(
                        template=(
                            "$column proportion of non-null values must be less than $max_value."
                        ),
                        params={
                            "column": {
                                "schema": {"type": "string"},
                                "value": "test_column",
                            },
                            "max_value": {
                                "schema": {"type": "number"},
                                "value": 0.8,
                            },
                            "strict_max": {
                                "schema": {"type": "boolean"},
                                "value": True,
                            },
                        },
                        schema={"type": "com.superconductive.rendered.string"},
                    ),
                    value_type="StringValueType",
                ),
            ],
            id="only_max_value_strict",
        ),
    ],
)
@pytest.mark.unit
def test_expectation_configuration_rendered_atomic_content(
    expectation: gxe.ExpectColumnProportionOfNonNullValuesToBeBetween,
    expected_expectation_configuration_rendered_atomic_content: list[RenderedAtomicContent],
):
    inline_renderer: InlineRenderer = InlineRenderer(render_object=expectation.configuration)

    actual_expectation_configuration_rendered_atomic_content: list[RenderedAtomicContent] = (
        inline_renderer.get_rendered_content()
    )

    assert len(actual_expectation_configuration_rendered_atomic_content) == 1

    assert (
        expected_expectation_configuration_rendered_atomic_content
        == actual_expectation_configuration_rendered_atomic_content
    )
